"""Usage of `setup.py` is deprecated, and is supplied only for legacy installation.
"""
import sys
import os
import os.path as osp
from pathlib import (
  Path,
  PurePath,
  PurePosixPath)
import importlib
import logging
import argparse
import subprocess
import tempfile
from argparse import RawTextHelpFormatter
logger = logging.getLogger(__name__)

#===============================================================================
def egg_info( args ):

  logger.warning(
    "running legacy 'setup.py egg_info'" )

  dir = Path(args.egg_base).joinpath(EGG_INFO_NAME)

  if not dir.exists():
    dir.mkdir(parents=True, exist_ok = True)

  with open(dir.joinpath('PKG-INFO'), 'wb' ) as fp:
    fp.write( PKG_INFO )

  with open( dir.joinpath('setup_requires.txt'), 'wb' ) as fp:
    fp.write( b'' )

  with open( dir.joinpath('requires.txt'), 'wb' ) as fp:
    fp.write( REQUIRES )

  with open( dir.joinpath('SOURCES.txt'), 'wb' ) as fp:
    fp.write( SOURCES )

  with open( dir.joinpath('top_level.txt'), 'wb' ) as fp:
    fp.write( TOP_LEVEL )

  with open( dir.joinpath('entry_points.txt'), 'wb' ) as fp:
    fp.write( ENTRY_POINTS )

  with open(dir.joinpath('dependency_links.txt'), 'wb' ) as fp:
    fp.write( b'' )

  with open( dir.joinpath('not-zip-safe'), 'wb' ) as fp:
    fp.write( b'' )

#===============================================================================
def bdist_wheel( args ):

  logger.warning(
    "running legacy 'setup.py bdist_wheel'" )

  sys.path = backend_path + sys.path

  backend = importlib.import_module( build_backend )

  backend.build_wheel(
    wheel_directory = args.dist_dir or args.bdist_dir or '.' )

#===============================================================================
def install( args ):

  logger.warning(
    "running legacy 'setup.py install'" )

  reqs = [ f"{{r}}" for r in build_requires ]

  subprocess.check_call([
    sys.executable,
    '-m',
    'pip',
    'install',
    *reqs ] )

  sys.path = backend_path + sys.path

  backend = importlib.import_module( build_backend )

  with tempfile.TemporaryDirectory() as tmpdir:
    wheel_name = backend.build_wheel(
      wheel_directory = tmpdir )

    subprocess.check_call([
      sys.executable,
      '-m',
      'pip',
      'install',
      tmpdir.joinpath(wheel_name) ])

#===============================================================================
def dummy( args ):
  pass

#===============================================================================
def main():

  logging.basicConfig(
    level = logging.INFO,
    format = "{{name}}:{{levelname}}: {{message}}",
    style = "{{" )


  logger.warning(
    "'setup.py' is deprecated, limited support for legacy installs. Upgrade pip." )

  parser = argparse.ArgumentParser(
    description = __doc__,
    formatter_class = RawTextHelpFormatter )

  subparsers = parser.add_subparsers()

  #.............................................................................
  egg_info_parser = subparsers.add_parser( 'egg_info' )

  egg_info_parser.set_defaults( func = egg_info )

  egg_info_parser.add_argument( "-e", "--egg-base",
    type = str,
    default = '.' )

  #.............................................................................
  bdist_wheel_parser = subparsers.add_parser( 'bdist_wheel' )

  bdist_wheel_parser.set_defaults( func = bdist_wheel )

  bdist_wheel_parser.add_argument( "-b", "--bdist-dir",
    type = str,
    default = '' )

  bdist_wheel_parser.add_argument( "-d", "--dist-dir",
    type = str,
    default = '' )

  bdist_wheel_parser.add_argument( "--python-tag",
    type = str,
    default = None )

  bdist_wheel_parser.add_argument( "--plat-name",
    type = str,
    default = None )

  bdist_wheel_parser.add_argument( "--py-limited-api",
    type = str,
    default = None )

  bdist_wheel_parser.add_argument( "--build-number",
    type = str,
    default = None )

  #.............................................................................
  install_parser = subparsers.add_parser( 'install' )

  install_parser.set_defaults( func = install )

  install_parser.add_argument( "--record",
    type = str,
    default = None )

  install_parser.add_argument( "--install-headers",
    type = str,
    default = None )

  install_parser.add_argument( "--compile",
    action='store_true' )

  install_parser.add_argument( "--single-version-externally-managed",
    action='store_true' )

  #.............................................................................
  clean_parser = subparsers.add_parser( 'clean' )

  clean_parser.set_defaults( func = dummy )

  clean_parser.add_argument( "-a", "--all",
    action='store_true' )

  args = parser.parse_args( )

  args.func( args )


#===============================================================================
# NOTE: these are templated literal values substituded by the backend when
# building the source distribution

build_backend = {build_backend}
backend_path = {backend_path}
build_requires = {build_requires}

EGG_INFO_NAME = {egg_info_name}

PKG_INFO = {pkg_info}

REQUIRES = {requires}

SOURCES = {sources}

TOP_LEVEL = {top_level}

ENTRY_POINTS = {entry_points}

#===============================================================================

if __name__ == "__main__":
  exit( main() )



from __future__ import annotations
import os
import glob
from pathlib import (
  Path)
import logging
from ..validate import (
  FileOutsideRootError,
  validating )
from ..path import (
  PathFilter,
  subdir,
  combine_ignore_patterns,
  resolve)

# #===============================================================================
# def rematch_replace(rematch, replace, name):

#   m = rematch.fullmatch(path)
#   if not m:
#     continue

#   args = (m.group(0), *m.groups())
#   kwargs = m.groupdict()
#   try:
#     replace.format(*args, **kwargs))
#   except (IndexError, KeyError) as e:
#     raise ValueError(
#       f"Replacement '{replace}' failed to format match '{m.group(0)}': "
#       f"{args}, {kwargs}") from None



#===============================================================================
def dist_iter(*,
  include,
  ignore,
  root ):

  patterns = PathFilter(
    patterns = ignore )

  for i, incl in enumerate(include):
    src = incl.src
    dst = incl.dst
    _ignore = incl.ignore

    _ignore_patterns = combine_ignore_patterns(
      patterns,
      PathFilter(
        patterns = _ignore,
        start = src ) )

    if not incl.include:
      yield ( i, src, dst, _ignore_patterns, True )
    else:
      for incl_pattern in incl.include:
        cwd = Path.cwd()
        try:
          # TODO: better way of globing *relative* to src directory
          os.chdir(src)
          matches = glob.glob(incl_pattern.glob, recursive = True)
        finally:
          os.chdir(cwd)

        for match in matches:
          match = Path(match)
          basename = match.parent
          src_filename = match.name

          m = incl_pattern.rematch.fullmatch(src_filename)
          if not m:
            continue

          args = (m.group(0), *m.groups())
          kwargs = m.groupdict()
          try:
            dst_filename = incl_pattern.replace.format(*args, **kwargs)
          except (IndexError, KeyError) as e:
            raise ValueError(
              f"Replacement '{incl_pattern.replace}' failed for"
              f" '{incl_pattern.rematch.pattern}':"
              f" {args}, {kwargs}") from None

          _src = src/basename/src_filename
          # re-base the dst path, path relative to src == path relative to dst
          _dst = dst/basename/dst_filename

          yield (i, _src, _dst, _ignore_patterns, False)


#===============================================================================
def dist_copy(*,
  base_path,
  include,
  ignore,
  dist,
  root = None,
  logger = None ):

  if len(include) == 0:
    return

  logger = logger or logging.getLogger( __name__ )

  with validating(key = 'copy'):

    for i, src, dst, ignore_patterns, individual in dist_iter(
      include = include,
      ignore = ignore,
      root = root ):

      with validating(key = i):

        dst = base_path.joinpath(dst)

        if not individual and ignore_patterns( src.parent, [src.name]):
          logger.debug( f'ignoring: {src}' )
          continue

        src_abs = resolve(src)

        if root and not subdir(root, src_abs, check = False):
          raise FileOutsideRootError(
            f"Must have common path with root:\n  file = \"{src_abs}\"\n  root = \"{root}\"")

        logger.debug(f"dist copy: {src} -> {dst}")

        if src.is_dir():
          dist.copytree(
            src = src,
            dst = dst,
            ignore = ignore_patterns )

        else:
          dist.copyfile(
            src = src,
            dst = dst )



from __future__ import annotations
from pathlib import (
  PurePosixPath)
from ..pep import (
  norm_dist_filename )
from ..pkginfo import PkgInfo
from .dist_base import dist_base
from .dist_targz import dist_targz

#===============================================================================
class dist_source_targz( dist_targz ):
  """Build a source distribution ``*.tar.gz`` file

  Parameters
  ----------
  pkg_info : :class:`PkgInfo <partis.pyproj.pkginfo.PkgInfo>`
  outdir : None | str | pathlib.Path
    Path to directory where the wheel file should be copied after completing build.
  tmpdir : None | str | pathlib.Path
    If not None, uses the given directory to place the temporary wheel file before
    copying to final location.
    My be the same as outdir.
  logger : None | :class:`logging.Logger`
    Logger to use.

  Example
  -------

  .. code-block:: python

    import tempfile

    with tempfile.TemporaryDirectory() as tmpdir:

      import os
      import os.path

      pkg_dir = os.path.join( tmpdir, 'src', 'my_package' )
      out_dir = os.path.join( tmpdir, 'build' )

      os.makedirs( pkg_dir )

      with open( os.path.join( pkg_dir, 'module.py' ), 'w' ) as fp:
        fp.write("print('hello')")


      from partis.pyproj import (
        PkgInfo,
        dist_source_targz )

      pkg_info = PkgInfo(
        project = dict(
          name = 'my-package',
          version = '1.0' ) )

      with dist_source_targz(
        pkg_info = pkg_info,
        outdir = out_dir ) as sdist:

        sdist.copytree(
          src = './src',
          dst = os.path.join( sdist.base_path, 'src' ) )


  """
  #-----------------------------------------------------------------------------
  def __init__( self,
    pkg_info,
    outdir = None,
    tmpdir = None,
    logger = None ):

    if not isinstance( pkg_info, PkgInfo ):
      raise ValueError(f"pkg_info must be instance of PkgInfo: {pkg_info}")

    self.pkg_info = pkg_info

    sdist_name_parts = [
      norm_dist_filename( self.pkg_info.name_normed ),
      norm_dist_filename( self.pkg_info.version ) ]

    self.base_path = PurePosixPath('-'.join( sdist_name_parts ))

    self.metadata_path = self.base_path.joinpath('PKG-INFO')

    super().__init__(
      outname = '-'.join( sdist_name_parts ) + '.tar.gz',
      outdir = outdir,
      tmpdir = tmpdir,
      logger = logger,
      named_dirs = {
        'root' : self.base_path,
        'metadata' : self.metadata_path } )


  #-----------------------------------------------------------------------------
  def finalize( self ):

    self.write(
      dst = self.metadata_path,
      data = self.pkg_info.encode_pkg_info() )


#===============================================================================
class dist_source_dummy( dist_base ):
  """Build a dummy source distribution without a physical file
  """

  #-----------------------------------------------------------------------------
  def __init__( self,
    pkg_info,
    outdir = None,
    tmpdir = None,
    logger = None ):

    if not isinstance( pkg_info, PkgInfo ):
      raise ValueError(f"pkg_info must be instance of PkgInfo: {pkg_info}")

    self.pkg_info = pkg_info

    sdist_name_parts = [
      self.pkg_info.name_normed,
      self.pkg_info.version ]

    self.base_path = PurePosixPath('-'.join( sdist_name_parts ))

    self.metadata_path = self.base_path.joinpath('PKG-INFO')

    super().__init__(
      outname = '-'.join( sdist_name_parts ) + '.tar.gz',
      outdir = outdir,
      tmpdir = tmpdir,
      logger = logger,
      named_dirs = {
        'root' : self.base_path,
        'metadata' : self.metadata_path } )


  #-----------------------------------------------------------------------------
  def create_distfile( self ): # pragma: no cover
    pass

  #-----------------------------------------------------------------------------
  def close_distfile( self ): # pragma: no cover
    pass

  #-----------------------------------------------------------------------------
  def copy_distfile( self ): # pragma: no cover
    pass

  #-----------------------------------------------------------------------------
  def remove_distfile( self ): # pragma: no cover
    pass

  #-----------------------------------------------------------------------------
  def finalize( self ): # pragma: no cover
    pass



from __future__ import annotations
import os
import os.path as osp
import sys
import shutil
from logging import (
  getLogger,
  Logger)
import tempfile
from copy import copy, deepcopy
from collections.abc import (
  Mapping,
  Sequence )
import subprocess
import multiprocessing
import glob
import warnings
import tomli
from pathlib import (
  Path,
  PurePath,
  PurePosixPath)

try:
  from importlib.metadata import metadata

except ImportError:
  from importlib_metadata import metadata

from .pkginfo import (
  PkgInfoReq,
  PkgInfo )
from .validate import (
  ValidationWarning,
  ValidationError,
  FileOutsideRootError,
  RequiredValueError,
  valid_dict,
  validating,
  valid,
  restrict,
  mapget )
from .norms import (
  scalar_list,
  norm_bool,
  norm_path_to_os,
  norm_path )
from .pep import (
  purelib_compat_tags,
  platlib_compat_tags )
from .path import (
  resolve)
from .load_module import (
  EntryPointError,
  EntryPoint,
  load_module,
  load_entrypoint )

from .legacy import legacy_setup_content

from .pptoml import (
  pptoml,
  # NOTE: deprecated
  pyproj_meson )

from .builder import (
  Builder )

from .dist_file import (
  dist_copy )

#===============================================================================
class PyProjBase:
  """Minimal build system for a Python project

  Extends beyond :pep:`517` and :pep:`621`


  Parameters
  ----------
  root :
    Path to the root project directory containing 'pyproject.toml'.
  logger :
    Parent logger to use when processing project.

  """
  #-----------------------------------------------------------------------------
  def __init__( self, *,
    root: Path,
    config_settings: dict|None = None,
    logger: Logger|None = None ):

    self.logger = logger or getLogger( __name__ )

    self.root = resolve(Path(root))

    self.pptoml_file = self.root / 'pyproject.toml'

    with open( self.pptoml_file, 'rb' ) as fp:
      src = fp.read()
      src = src.decode( 'utf-8', errors = 'replace' )
      self._pptoml = tomli.loads( src )

    with validating(root = self._pptoml, file = self.pptoml_file):
      self._pptoml = pptoml(self._pptoml)

      with validating(key = 'tool'):
        if 'tool' not in self.pptoml:
          # TODO: !!!
          raise RequiredValueError("tool.pyproj is required for backend")

        with validating(key = 'pyproj'):
          if 'pyproj' not in self.pptoml.tool:
            raise RequiredValueError("tool.pyproj is required for backend")

      if self.project.dynamic and 'prep' not in self.pyproj:
        raise RequiredValueError("tool.pyproj.prep is required to resolve project.dynamic")

    #...........................................................................
    # construct a validator from the tool.pyproj.config table
    # NOTE: only really used in the event that "config_settings" are passed to the backend
    config_default = dict()

    for k,v in self.pyproj.config.items():
      if isinstance(v, bool):
        config_default[k] = valid(v, norm_bool)

      elif isinstance(v, scalar_list):
        config_default[k] = restrict(*v)

      else:
        config_default[k] = valid(v, type(v))

    class valid_config_settings(valid_dict):
      allow_keys = list()
      default = config_default

    with validating( key = 'config_settings' ):
      self._config_settings = valid_config_settings(config_settings or dict())

    #...........................................................................
    self.build_backend = mapget( self.pptoml,
      'build-system.build-backend',
      "" )

    self.backend_path = mapget( self.pptoml,
      'build-system.backend-path',
      list() )

    #...........................................................................
    # default build requirements
    self.build_requires = self.pptoml.build_system.requires

    #...........................................................................
    # used to create name for binary distribution
    self.is_platlib = bool(self.binary.platlib.copy)

    if self.is_platlib:
      self.binary.compat_tags = platlib_compat_tags()

    #...........................................................................
    self.prep()

    with validating(
      key = 'project',
      root = self._pptoml,
      file = self.pptoml_file):

      self.pkg_info = PkgInfo(
        project = self.project,
        root = self.root )

    # Update logger once package info is created
    self.logger = self.logger.getChild( f"['{self.pkg_info.name_normed}']" )

  #-----------------------------------------------------------------------------
  @property
  def pptoml(self):
    """pptoml : Parsed and validated pyproject.toml document
    """
    return self._pptoml

  #-----------------------------------------------------------------------------
  @property
  def project(self):
    """:class:`partis.pyproj.pptoml.project`
    """
    return self._pptoml.project

  #-----------------------------------------------------------------------------
  @property
  def pyproj(self):
    """:class:`partis.pyproj.pptoml.pyproj`
    """
    return self._pptoml.tool.pyproj

  #-----------------------------------------------------------------------------
  @property
  def config_settings(self):
    """Config settings passed to backend, or defaults from ``pyproj.config``
    """
    return self._config_settings

  #-----------------------------------------------------------------------------
  # alias for backward compatibility
  config = config_settings

  #-----------------------------------------------------------------------------
  @property
  def targets(self):
    """
    """
    return self._pptoml.tool.pyproj.targets

  #-----------------------------------------------------------------------------
  @property
  def meson(self):
    """:class:`partis.pyproj.pptoml.pyproj_meson`

    .. deprecated:: 0.1.0
      Use :attr:`PyProjBase.targets` to access all build targets.
      These are no longer restricted to meson, but this attribute kept for backward
      compatability.

      Inplace changes to the returned object are not propagated back to the target
      configuration.

    """
    targets = self._pptoml.tool.pyproj.targets

    if not (len(targets) == 1 and targets[0].entry == 'partis.pyproj.builder:meson'):
      raise ValidationError("The 'meson' attribute is undefined for targets")

    meson = dict(targets[0])
    meson.pop('entry')
    meson.pop('work_dir')
    meson.pop('env')
    meson['compile'] = meson.pop('enabled')
    return pyproj_meson(meson)

  #-----------------------------------------------------------------------------
  @property
  def dist(self):
    """:class:`partis.pyproj.pptoml.pyproj_dist`
    """
    return self._pptoml.tool.pyproj.dist

  #-----------------------------------------------------------------------------
  @property
  def source(self):
    """:class:`partis.pyproj.pptoml.pyproj_dist_source`
    """
    return self._pptoml.tool.pyproj.dist.source

  #-----------------------------------------------------------------------------
  @property
  def binary(self):
    """:class:`partis.pyproj.pptoml.pyproj_dist_binary`
    """
    return self._pptoml.tool.pyproj.dist.binary


  #-----------------------------------------------------------------------------
  @property
  def add_legacy_setup(self):
    """bool
    """
    return self.dist.source.add_legacy_setup

  #-----------------------------------------------------------------------------
  @property
  def build_requires(self):
    """set[:class:`PkgInfoReq`]
    """
    return self._build_requires

  #-----------------------------------------------------------------------------
  @build_requires.setter
  def build_requires(self, reqs):
    self._build_requires = set([ PkgInfoReq(r) for r in reqs ])

  #-----------------------------------------------------------------------------
  def prep_entrypoint( self, name, obj, logger ):

    prep = obj.get( 'prep', None )

    if not prep:
      return None

    entry_point = EntryPoint(
      pyproj = self,
      root = self.root,
      name = name,
      logger = logger,
      entry = prep.entry )

    entry_point(**prep.kwargs)

  #-----------------------------------------------------------------------------
  def prep( self ):
    """Prepares project metadata
    """
    # backup project to detect changes made by prep
    project = deepcopy(self.project)
    dynamic = project.dynamic

    self.prep_entrypoint(
      name = "tool.pyproj.prep",
      obj = self.pyproj,
      logger = self.logger.getChild("prep") )

    # NOTE: check that any dynamic meta-data is defined after prep
    for k in dynamic:
      # all dynamic keys should updated by prep
      if k not in self.project or self.project[k] == project[k]:
        warnings.warn(
          f"project.dynamic listed key as dynamic, but not altered in prep: {k}",
          ValidationWarning )

    for k, v in self.project.items():
      if k not in dynamic and (k not in project or project[k] != v):
        # don't allow keys to be added or changed unless they were listed in dynamic
        raise ValidationError(
          f"prep updated key not listed in project.dynamic: {k}" )

    # fields are no longer dynamic
    self.project.dynamic = list()

    # make sure build requirements are still a set of PkgInfoReq
    self.build_requires = self.build_requires

  #-----------------------------------------------------------------------------
  def dist_prep( self ):
    """Prepares project files for a distribution
    """

    self.prep_entrypoint(
      name = "tool.pyproj.dist.prep",
      obj = self.dist,
      logger = self.logger.getChild("dist.prep") )


  #-----------------------------------------------------------------------------
  def dist_source_prep( self ):
    """Prepares project files for a source distribution
    """

    self.prep_entrypoint(
      name = "tool.pyproj.dist.source.prep",
      obj = self.dist.source,
      logger = self.logger.getChild("dist.source.prep") )

  #-----------------------------------------------------------------------------
  def dist_source_copy( self, *, dist ):
    """Copies prepared files into a source distribution

    Parameters
    ---------
    sdist : :class:`dist_base <partis.pyproj.dist_file.dist_base.dist_base>`
      Builder used to write out source distribution files
    """

    with validating( key = 'tool.pyproj.dist.source'):
      dist_copy(
        base_path = dist.named_dirs['root'],
        include = self.source.copy,
        ignore = self.dist.ignore + self.source.ignore,
        dist = dist,
        root = self.root,
        logger = self.logger )

      if self.add_legacy_setup:
        with validating(key = 'add_legacy_setup'):

          self.logger.info("generating legacy 'setup.py'")
          legacy_setup_content( self, dist )

  #-----------------------------------------------------------------------------
  def dist_binary_prep( self ):
    """Prepares project files for a binary distribution
    """

    builder = Builder(
      pyproj = self,
      root = self.root,
      targets = self.targets,
      logger = self.logger.getChild("targets"))

    with builder:
      builder.build_targets()

      self.prep_entrypoint(
        name = "tool.pyproj.dist.binary.prep",
        obj = self.binary,
        logger = self.logger.getChild("dist.binary.prep") )

    self.logger.debug(f"Compatibility tags after dist.binary.prep: {self.binary.compat_tags}")

  #-----------------------------------------------------------------------------
  def dist_binary_copy( self, *, dist ):
    """Copies prepared files into a binary distribution

    Parameters
    ---------
    bdist : :class:`dist_base <partis.pyproj.dist_file.dist_base.dist_base>`
      Builder used to write out binary distribution files
    """


    with validating(key = 'tool.pyproj.dist.binary'):
      ignore = self.dist.ignore + self.dist.binary.ignore

      dist_copy(
        base_path = dist.named_dirs['root'],
        include = self.binary.copy,
        ignore = ignore,
        dist = dist,
        root = self.root,
        logger = self.logger )

      data_scheme = [
        'data',
        'headers',
        'scripts',
        'purelib',
        'platlib' ]

      for k in data_scheme:
        if k in self.binary:

          dist_data = self.binary[k]

          _include = dist_data.copy
          _ignore = ignore + dist_data.ignore

          with validating( key = k ):
            dist_copy(
              base_path = dist.named_dirs[k],
              include = _include,
              ignore = _ignore,
              dist = dist,
              root = self.root,
              logger = self.logger )



from __future__ import annotations
import sys
import warnings
from functools import partial
import inspect
import types
from copy import copy
from abc import ABCMeta
from collections.abc import (
  Mapping,
  Sequence,
  Iterable )

#===============================================================================
# NOTE: Filtering works by changing the traceback linked-list, which means
# writing to the 'tb_next' attrbute to the next frame not to be skipped.
# However, 'tb_next' was a read-only attribute until Python 3.7
FILTER_VALIDATING_FRAMES = sys.version_info >= (3,7)

#===============================================================================
def filter_traceback(traceback, ignore):
  # NOTE: always keep the first frame in the trace-back (even if it would be filtered)
  cur_tb = traceback
  prev_kept_tb = cur_tb

  # start filtering at the next inner frame
  cur_tb = cur_tb.tb_next

  while cur_tb is not None:
    frame = cur_tb.tb_frame
    lineno = cur_tb.tb_lineno
    code = frame.f_code

    cur_tb = cur_tb.tb_next

    # check the ignore function to see if this frame should be kept
    # NOTE: always keep the last frame in the trace-back (the initial 'raise')
    if cur_tb is not None and ignore(frame, lineno):
      # if ignored, set the 'tb_next' attribute of the previously kept frame
      # to the next frame to be checked
      prev_kept_tb.tb_next = cur_tb
    else:
      prev_kept_tb = cur_tb

#===============================================================================
def ignore_validating(frame, lineno):
  if frame.f_code.co_filename == __file__:
    return True

  return False

#===============================================================================
class ValidationWarning( RuntimeWarning ):
  pass

#===============================================================================
class ValidationError( ValueError ):
  """General validation error

  Parameters
  ----------
  msg : str
    Error message
  """
  def __init__( self, msg,
    doc_root = None,
    doc_file = None,
    doc_path = None,
    extra = None):

    msg = inspect.cleandoc( msg )

    self.msg = msg
    self.doc_root = doc_root
    self.doc_file = doc_file
    self.doc_path = doc_path or list()
    self.extra = extra

    super().__init__( msg )

  #-----------------------------------------------------------------------------
  def __str__(self):

    parts = list()

    if self.doc_path:
      _path = self.doc_path[0]

      if isinstance(_path, int):
        _path = f"[{_path}]"

      for k in self.doc_path[1:]:
        if isinstance(k, int):
          _path += f"[{k}]"
        else:
          _path += f".{k}"

      parts.append( f"at `{_path}`" )

    if self.doc_file:
      parts.append(f"in \"{self.doc_file}\"")


    loc = " ".join(parts)
    msg = self.msg

    if loc:
      msg += '\n' + loc

    if self.extra:
      msg += '\n' + str(self.extra)

    return msg

  #-----------------------------------------------------------------------------
  def __repr__(self):
    return str(self)

  #-----------------------------------------------------------------------------
  def model_hint(self):
    from partis.utils import ModelHint, Loc

    return ModelHint(
      msg = type(self).__name__,
      data = self.msg,
      level = 'error',
      loc = Loc(
        filename = self.doc_file,
        path = self.doc_path ))

#===============================================================================
class RequiredValueError( ValidationError ):
  pass

#===============================================================================
class ValidDefinitionError( ValidationError ):
  pass

#===============================================================================
class ValidPathError(ValidationError):
  """File is not valid
  """
  pass

#===============================================================================
class FileOutsideRootError(ValidPathError):
  """File path is outside a desired root directory
  """
  pass

#===============================================================================
class validating:
  """Context manager to append information to a ValidationError as it propagates

  Intermediate frames for internal validating routines, except for the first
  and last frame, are filtered out of any tracebacks.

  Parameters
  ----------
  key: None | str
    Insert the current key being validated to the head of the 'doc_path'
  root: None | Mapping | Sequence
    Set the root document being validated as 'doc_root'
  file: None | str
    Set a file as the source of the data being validated as 'doc_file'

  See Also
  --------
  * ValidationError
  """
  #-----------------------------------------------------------------------------
  def __init__(self,
    key = None,
    root = None,
    file = None):

    self.key = key
    self.root = root
    self.file = file

  #-----------------------------------------------------------------------------
  def __enter__(self):
    return self

  #-----------------------------------------------------------------------------
  def __exit__(self, type, value, traceback):
    if type is not None:
      if issubclass(type, ValidationError):

        if FILTER_VALIDATING_FRAMES:
          filter_traceback(traceback, ignore_validating)

        value.doc_root = value.doc_root or self.root
        value.doc_file = value.doc_file or self.file

        if self.key is not None:
          value.doc_path.insert(0, self.key)

      else:
        raise ValidationError(
          "Error while validating",
          doc_root = self.root,
          doc_file = self.file,
          doc_path = None if self.key is None else [self.key] ) from value

    # do not handle any exceptions here
    return False

#===============================================================================
class Special:
  #-----------------------------------------------------------------------------
  def __repr__(self):
    return type(self).__name__

  #-----------------------------------------------------------------------------
  def __eq__(self, other):
    return type(self) is type(other)

  #-----------------------------------------------------------------------------
  def __ne__(self, other):
    return type(self) is not type(other)

  #-----------------------------------------------------------------------------
  def __hash__(self):
    return hash(str(self))

#===============================================================================
class Optional(Special):
  """Optional value
  """
  pass

#===============================================================================
class OptionalNone(Special):
  """Optional value, but is set to None if not initially set
  """
  pass

#===============================================================================
class Required(Special):
  """Required value
  """
  pass

#===============================================================================
class NotSet(Special):
  """Special value indicating a value is not set
  """
  pass

OPTIONAL = Optional()
OPTIONAL_NONE = OptionalNone()
REQUIRED = Required()
NOTSET = NotSet()

#===============================================================================
def validate(val, default, validators):
  """Internal method to apply default value and validators
  """

  if val in [NOTSET, None]:
    if default in [OPTIONAL, OPTIONAL_NONE]:
      #               | NOTSET | None |
      # --------------+--------+------+
      # OPTIONAL      | NOTSET | None |
      # --------------+--------+------+
      # OPTIONAL_NONE | None   | None |
      return None if default == OPTIONAL_NONE else val

    elif default == REQUIRED:
      raise RequiredValueError("Value is required")

    else:
      val = default

  if not isinstance(validators, Sequence):
    validators = [validators]

  for validator in validators:
    if isinstance(validator, type):
      # cast to valid type (if needed)
      if not isinstance(val, validator):
        try:
          val = validator(val)
        except ValidationError as e:
          # already a validation error
          raise e

        except Exception as e:
          # re-raise other errors as a ValidationError
          raise ValidationError(
            f"Failed to cast type {type(val).__name__} to {validator.__name__}: {val}") from e

    elif callable(validator):
      val = validator(val)

    elif isinstance(validator, Sequence):
      # union (only one needs to succeed)
      errs = list()

      for _validator in validator:
        try:
          val = validate(val, REQUIRED, _validator)
          break
        except Exception as e:
          errs.append((_validator, e))

      else:
        if errs:
          errs = '\n'.join([f"- {k} -> {v}" for k,v in errs])
          raise ValidationError(f"Value must pass at least one of the validators:\n{errs}")

  return val

#===============================================================================
def fmt_validator(v):

  if isinstance(v, Validator):
    return str(v)

  if v == '':
    return "''"

  if not (
    callable(v)
    or any(
      isinstance(v,t)
      for t in [type, types.BuiltinFunctionType, types.FunctionType]) ):

    return str(v)

  name = None

  while name is None:
    if hasattr(v, '__qualname__'):
      name = v.__qualname__

    elif hasattr(v, '__name__'):
      name = v.__name__

    else:
      v = type(v)

  if name.startswith('<'):
    return name

  mod = None

  if hasattr(v, '__module__'):
    mod = v.__module__

    if mod != 'builtins':
      name = f"{mod}.{name}"

  return f"<{name}>"

#===============================================================================
class Validator:
  """Validates a value
  """
  __slots__ = ('default', 'validators')

  #-----------------------------------------------------------------------------
  def __init__(self, *args, default = NOTSET):

    args = list(args)

    # TODO: change comparisons from 'is' to "==", reimplement Special __eq__
    if default == NOTSET:
      default = REQUIRED

      if len(args):
        v = args.pop(0)

        if v in [REQUIRED, OPTIONAL, OPTIONAL_NONE]:
          # set from special value
          default = v

        elif v is None:
          # set to fill in None if not given
          default = OPTIONAL_NONE

        elif any(isinstance(v, t) for t in [bool, int, float, str, Sequence, Mapping]):
          default = v

        elif isinstance(v, type):
          # still used to validate the type
          args.insert(0, v)

          try:
            default = v()
          except RequiredValueError:
            # this really is a required value
            pass

          except Exception as e:
            # the type cannot be instantiated without arguments
            raise ValidDefinitionError(
              "Default value must be specified, or explicitly set as optional or required") from e

        else:
          # cannot be used as default, put back to use as validator
          # and a value is assumed to be required
          args.insert(0, v)

    elif default is None:
      # set to fill in None if not given
      default = OPTIONAL_NONE

    if len(args) == 0 and default not in [REQUIRED, OPTIONAL, OPTIONAL_NONE]:
      # convenience method to used default value to derive type
      args.append(type(default))

    self.default = default
    self.validators = tuple(args)

  #-----------------------------------------------------------------------------
  def __str__(self):
    args = list()
    for v in self.validators:
      args.append(fmt_validator(v))

    args.append(f"default = {fmt_validator(self.default)}")
    args = ', '.join(args)
    return f"{type(self).__name__}({args})"

  #-----------------------------------------------------------------------------
  def __repr__(self):
    return str(self)

  #-----------------------------------------------------------------------------
  def __call__(self, val = NOTSET):
    return validate(val, self.default, self.validators)

#===============================================================================
class Restricted(Validator):
  """Restricts a value to one of listed options
  """
  __slots__ = ('options',)

  #-----------------------------------------------------------------------------
  def __init__(self, *options ):
    if len(options) == 0:
      raise ValidDefinitionError("Must have at least one option")

    super().__init__(options[0], type(options[0]))

    _options = list()

    with validating(key = 'options'):
      for i,v in enumerate(options):
        with validating(key = i):
          _options.append(super().__call__(v))

    self.options = set(_options)

  #-----------------------------------------------------------------------------
  def __call__(self, val):
    val = super().__call__(val)

    if val not in self.options:
      raise ValidationError(
        f"Must be one of {self.options}: {val}")

    return val

#===============================================================================
def valid(*validators, default = NOTSET):
  """Casts list of objects to Validator, if needed
  """
  if len(validators) == 1:
    v = validators[0]

    if isinstance(v, Validator):
      return v

  return Validator(*validators, default = default)

#===============================================================================
def union(*validators):
  """Value must pass at least one of listed validators
  """
  return Validator([valid(v) for v in validators], default = REQUIRED)

#===============================================================================
def restrict(*options):
  """Restricts a value to one of listed options
  """
  return Restricted(*options)

#===============================================================================
def valid_type(
  obj,
  types ):

  for t in types:
    if isinstance( obj, t ):
      return t

  raise ValidationError(
    f"Must be of type {types}: {type(obj)}" )

#===============================================================================
def valid_keys(
  obj,
  key_valid = None,
  value_valid = None,
  item_valid = None,
  allow_keys = None,
  require_keys = None,
  min_keys = None,
  wedge_keys = None,
  mutex_keys = None,
  deprecate_keys = None,
  forbid_keys = None,
  default = None,
  proxy_keys = None ):
  """Check that a mapping does not contain un-expected keys

  Parameters
  ----------
  obj: Mapping
    Mapping object to validate
  key_valid: None | callable
    Validates all keys
  value_valid: None | callable
    Validates all values
  item_valid: None | callable
    Validates all (key,value) pairs
  allow_keys: None | list[str]
    Mapping may not contain keys that are not listed.
  require_keys: None | list[str]
    Mapping must contain all listed keys.
  min_keys: None | list[ list[str] ]
    Mapping must contain at least one key from each list.
  wedge_keys: None | list[ list[str] ]
    Mapping must contain either none or all of the listed keys.
  mutex_keys: None | list[ list[str] ]
    Mapping may contain at most one key from each list.
  deprecate_keys: None | list[ (str, None | str | Required) ]
    First key is marked as deprecated and removed from the Mapping.
    If new key is given, the value is remapped to the new key.
    If new key is Required, an error is raised, otherwise a deprecation warning
    is reported.
  forbid_keys: None | list[str]
    Mapping must not contain any of the listed keys.
  default: None | Mapping[object, object | type | Validator]
    Default value or validator for given keys.
  proxy_keys : None | list[ list[str] ]
  """

  if not isinstance( obj, Mapping ):
    raise ValidationError(
      f"Must be mapping: {type(obj)}" )

  out = obj

  def copy_once(out):
    if out is obj:
      # copy only if 'out' is still the same object as obj
      return copy(obj)

    return out

  if forbid_keys:
    for k in forbid_keys:
      if k in out:
        raise ValidationError(f"Use of key '{k}' is not allowed")

  if deprecate_keys:
    for k_old, k_new in deprecate_keys:
      if k_old in out:
        out = copy_once(out)

        if k_new and k_new not in [OPTIONAL, OPTIONAL_NONE]:
          if k_new == REQUIRED:
            # Use of REQUIRED indicates this is an error
            raise ValidationError(f"Use of key '{k_old}' is deprecated")
          else:
            warnings.warn(f"Use of key '{k_old}' is deprecated, replaced by '{k_new}'", DeprecationWarning)

          if k_new not in out:
            out[k_new] = out[k_old]

        else:
          warnings.warn(f"Use of key '{k_old}' is deprecated", DeprecationWarning)

        out.pop(k_old)

  if proxy_keys:
    for k, k_src in proxy_keys:
      if k_src in out and out.get(k, None) is None:
        out = copy_once(out)
        out[k] = out[k_src]

  if key_valid:
    for k, v in out.items():
      with validating(key = k):
        _k = key_valid(k)

      if _k != k:
        out = copy_once(out)
        out.pop(k)
        out[_k] = v

  if value_valid:
    for k, v in out.items():
      with validating(key = k):
        _v = value_valid(v)

      if _v is not v:
        out = copy_once(out)
        out[k] = _v

  if item_valid:
    for k,v in out.items():
      with validating(key = k):
        _k, _v = item_valid( (k,v) )

      if _k != k:
        out = copy_once(out)
        out.pop(k)
        out[_k] = _v

      elif _v is not v:
        out = copy_once(out)
        out[k] = _v


  if allow_keys is not None:
    allow_keys = copy(allow_keys)
    allow_keys.extend( require_keys or [] )
    allow_keys.extend( [
        k_new
        for k_old, k_new in deprecate_keys
        if k_new not in [None, OPTIONAL, OPTIONAL_NONE, REQUIRED ] ]
      if deprecate_keys else [] )

    allow_keys.extend( default.keys() if default else [] )

    if proxy_keys:
      for keys in proxy_keys:
        allow_keys.extend(keys)

    if min_keys:
      for keys in min_keys:
        allow_keys.extend(keys)

    if wedge_keys:
      for keys in wedge_keys:
        allow_keys.extend(keys)

    if mutex_keys:
      for keys in mutex_keys:
        allow_keys.extend(keys)

    allow_keys = set(allow_keys)

    for k in out.keys():
      if k not in allow_keys:
        raise ValidationError(
          f"Allowed keys {allow_keys}: '{k}'" )

  if default:
    for k, v in default.items():
      if not isinstance(v, Validator):
        # convert literal value to a validator
        v = valid(v)

      val = out.get(k, NOTSET)

      with validating(key = k):
        _val = v( val )

      if _val is not val:
        out = copy_once(out)
        out[k] = _val

  if require_keys:
    for k in require_keys:
      if k not in out:
        raise ValidationError(
          f"Required keys {require_keys}: '{k}'" )

  if min_keys:
    for keys in min_keys:
      if not any(k in out for k in keys):
        raise ValidationError(
          f"Must have at least one of keys: {keys}" )

  if wedge_keys:
    for keys in wedge_keys:
      if any(k in out for k in keys) and not all(k in out for k in keys):
        raise ValidationError(
          f"Must have either all, or none, of keys: {keys}" )

  if mutex_keys:
    for keys in mutex_keys:
      if sum(k in out for k in keys) > 1:
        raise ValidationError(
          f"May not have more than one of keys: {keys}" )


  return out

#===============================================================================
class _ValidDictAttr:
  r"""Descriptor for accessing dictionary keys as attributes
  """
  __slots__ = ('name', 'key')

  #-----------------------------------------------------------------------------
  def __init__(self, name: str, key: str):
    self.name = name
    self.key = key

  #-----------------------------------------------------------------------------
  def __get__(self, obj, owner):
    if obj is None:
      return self

    return obj._data[self.key]

  #-----------------------------------------------------------------------------
  def __set__(self, obj, value):
    obj._data[self.key] = value
    obj._validate()

  #-----------------------------------------------------------------------------
  def __delete__(self, obj):
    del obj._data[self.key]
    obj._validate()

  #-----------------------------------------------------------------------------
  def __repr__(self):
    return f"{type(self).__name__}({self.name!r}, {self.key!r})"

#===============================================================================
class _ValidDictMeta(ABCMeta):
  #-----------------------------------------------------------------------------
  def __new__(mcls,
      name,
      bases,
      namespace):

    schema_keys = (
      'key_valid',
      'value_valid',
      'item_valid',
      'allow_keys',
      'require_keys',
      'min_keys',
      'wedge_keys',
      'mutex_keys',
      'deprecate_keys',
      'forbid_keys',
      'default',
      'proxy_keys')

    schema = {}

    for base in bases:
      # include options from base classes
      if isinstance(base, _ValidDictMeta):
        for k in schema_keys:
          v = base.__dict__.get(k)

          if v is not None:
            schema[k] = v

    for k in schema_keys:
      v = namespace.get(k)

      if v is not None:
        schema[k] = v

    namespace.update(schema)

    validator = schema.get('validator')
    key_valid = schema.get('key_valid')
    value_valid = schema.get('value_valid')
    item_valid = schema.get('item_valid')
    allow_keys = schema.get('allow_keys')
    require_keys = schema.get('require_keys')
    min_keys = schema.get('min_keys')
    wedge_keys = schema.get('wedge_keys')
    mutex_keys = schema.get('mutex_keys')
    deprecate_keys = schema.get('deprecate_keys')
    forbid_keys = schema.get('forbid_keys')
    default = schema.get('default')
    proxy_keys = schema.get('proxy_keys')

    default = {
      k: valid(v)
      for k,v in (default or dict()).items()}
    all_keys = list()

    all_keys.extend(allow_keys or ())
    all_keys.extend(require_keys or ())
    all_keys.extend(default.keys())

    if deprecate_keys:
      all_keys.extend([
          k_new
          for k_old, k_new in deprecate_keys
          if k_new not in [None, OPTIONAL, OPTIONAL_NONE, REQUIRED]])

    if min_keys:
      for keys in min_keys:
        all_keys.extend(keys)

    if wedge_keys:
      for keys in wedge_keys:
        all_keys.extend(keys)

    if mutex_keys:
      for keys in mutex_keys:
        all_keys.extend(keys)

    all_keys = tuple(all_keys)

    # compose validator with parameterized valid_keys method
    validator = valid(
      validator or (lambda v: v),
      partial(valid_keys,
        key_valid = key_valid,
        value_valid = value_valid,
        item_valid = item_valid,
        allow_keys = allow_keys,
        require_keys = require_keys,
        min_keys = min_keys,
        wedge_keys = wedge_keys,
        mutex_keys = mutex_keys,
        deprecate_keys = deprecate_keys,
        forbid_keys = forbid_keys,
        default = default,
        proxy_keys = proxy_keys))

    namespace['_all_keys'] = all_keys
    namespace['_validator'] = validator
    namespace.setdefault('__slots__', ())

    for k in all_keys:
      # create attributes mapping to dictionary keys
      name = k.replace('-','_')
      attr = _ValidDictAttr(name, k)
      namespace[name] = attr

    cls = super().__new__(mcls, name, bases, namespace)

    return cls

#===============================================================================
class valid_dict(Mapping, metaclass = _ValidDictMeta):
  """Validated Mapping

  Attributes
  ----------
  proxy_key: None | str
    If initialized with a value that is not a Mapping, this key is assigned the
    value before performing validation.
  key_valid: None | callable
    Validates all keys
  value_valid: None | callable
    Validates all values
  item_valid: None | callable
    Validates all (key,value) pairs
  allow_keys: None | list[str]
    Mapping may not contain keys that are not listed.
  require_keys: None | list[str]
    Mapping must contain all listed keys.
  min_keys: None | list[ list[str] ]
    Mapping must contain at least one key from each list.
  wedge_keys: None | list[ list[str] ]
    Mapping must contain either none or all of the listed keys.
  mutex_keys: None | list[ list[str] ]
    Mapping may contain at most one key from each list.
  deprecate_keys: None | list[ (str, None | str | Required) ]
    First key is marked as deprecated and removed from the Mapping.
    If new key is given, the value is remapped to the new key.
    If new key is Required, an error is raised, otherwise a deprecation warning
    is reported.
  forbid_keys: None | list[str]
    Mapping must not contain any of the listed keys.
  default: None | Mapping[object, object | type | Validator]
    Default value or validator for given keys.
  validator : None | Validator
    General validator for entire Mapping after above constraints are satisfied.
  See Also
  --------
  * :func:`valid_keys`
  """
  __slots__ = ('_data',)

  proxy_key = None
  proxy_keys = None
  key_valid = None
  value_valid = None
  item_valid = None
  allow_keys = None
  require_keys = None
  min_keys = None
  wedge_keys = None
  mutex_keys = None
  deprecate_keys = None
  forbid_keys = None
  default = None
  validator = None

  #---------------------------------------------------------------------------#
  # pylint: disable-next=E0602
  def __init__( self, *args, **kwargs ):
    cls = type(self)

    if (
      len(kwargs) == 0
      and len(args) == 1
      and not isinstance(args[0], Mapping) ):

      v = args[0]

      if v in (None, OPTIONAL, OPTIONAL_NONE):
        args = [dict()]

      elif cls.proxy_key:
        args = [{ cls.proxy_key : v }]

    self._data = dict(*args, **kwargs)
    self._validate()

  #-----------------------------------------------------------------------------
  def __eq__(self, other):
    if isinstance(other, valid_dict):
      other = other._data

    return self._data.__eq__(other)

  #-----------------------------------------------------------------------------
  def __ne__(self, other):
    if isinstance(other, valid_dict):
      other = other._data

    return self._data.__ne__(other)

  #-----------------------------------------------------------------------------
  def __copy__(self):
    cls = type(self)
    obj = cls.__new__(cls)
    obj._data = copy(self._data)
    return obj

  #-----------------------------------------------------------------------------
  def __str__(self):
    return str(self._data)

  #-----------------------------------------------------------------------------
  def __repr__(self):
    return repr(self._data)

  #-----------------------------------------------------------------------------
  def __len__(self):
    return len(self._data)

  #-----------------------------------------------------------------------------
  def __iter__(self):
    return iter(self._data)

  #-----------------------------------------------------------------------------
  def __contains__(self, name):
    return self._data.__contains__(name)

  #-----------------------------------------------------------------------------
  def keys(self):
    return self._data.keys()

  #-----------------------------------------------------------------------------
  def values(self):
    return self._data.values()

  #-----------------------------------------------------------------------------
  def items(self):
    return self._data.items()

  #-----------------------------------------------------------------------------
  def clear(self):
    self._data.clear()
    self._validate()

  #---------------------------------------------------------------------------#
  def update(self, *args, **kwargs ):
    self._data.update(*args, **kwargs)
    self._validate()

  #-----------------------------------------------------------------------------
  def setdefault( self, *args, **kwargs ):
    val = self._data.setdefault(*args, **kwargs)
    self._validate()
    return val

  #-----------------------------------------------------------------------------
  def get( self, *args, **kwargs ):
    return self._data.get(*args, **kwargs)

  #-----------------------------------------------------------------------------
  def pop( self, *args, **kwargs ):
    val = self._data.pop(*args, **kwargs)
    self._validate()
    return val

  #-----------------------------------------------------------------------------
  def popitem( self, *args, **kwargs ):
    val = self._data.popitem(*args, **kwargs)
    self._validate()
    return val

  #-----------------------------------------------------------------------------
  def __getitem__( self, key ):
    return self._data.__getitem__(key)

  #-----------------------------------------------------------------------------
  def __setitem__( self, key, val ):
    self._data.__setitem__(key, val)
    self._validate()

  #-----------------------------------------------------------------------------
  def __delitem__( self, key ):
    self._data.__delitem__( key )
    self._validate()

  #-----------------------------------------------------------------------------
  def _validate(self):
    self._data = self._validator(self._data)

#===============================================================================
class valid_list(list):
  """Validated list
  """
  _as_list = None
  value_valid = None
  _min_len = 0

  #---------------------------------------------------------------------------#
  def __init__( self, vals = None ):
    cls = type(self)
    self._as_list = cls._as_list or list
    self.value_valid = valid(
      cls.value_valid or (lambda v: v))

    if vals is None:
      vals = list()

    elif self._as_list:
      vals = self._as_list(vals)

    for i,v in enumerate(vals):
      with validating(key = i):
        vals[i] = self.value_valid(v)

    super().__init__(vals)
    self._validate()

  #-----------------------------------------------------------------------------
  def clear(self):
    super().clear()
    self._validate()

  #-----------------------------------------------------------------------------
  def pop( self, *args, **kwargs ):
    val = super().pop(*args, **kwargs)
    self._validate()
    return val

  #---------------------------------------------------------------------------#
  def append(self, val ):
    with validating(key = len(self)):
      val = self.value_valid(val)

    super().append(val)

  #---------------------------------------------------------------------------#
  def extend(self, vals ):
    vals = list(vals)

    for i,v in enumerate(vals):
      with validating(key = len(self) + i):
        vals[i] = self.value_valid(v)

    super().extend(vals)

  #-----------------------------------------------------------------------------
  def __setitem__( self, key, val ):
    with validating(key = key):
      val = self.value_valid(val)

    super().__setitem__(key, val)

  #-----------------------------------------------------------------------------
  def _validate(self):
    if len(self) < self._min_len:
      raise ValidationError(f"Must have length >= {self._min_len}: {len(self)}")

#===============================================================================
def mapget(
  obj,
  path,
  default = None ):
  """Convenience method for extracting a value from a nested mapping
  """

  parts = path.split('.')
  _obj = obj
  last_i = len(parts)-1

  for i, part in enumerate(parts):
    if not isinstance( _obj, Mapping ):
      lpath = '.'.join(parts[:i])
      rpath = '.'.join(parts[i:])

      if len(lpath) > 0:
        raise ValidationError(
          f"Expected a mapping object [{lpath}][{rpath}]: {_obj}")
      else:
        raise ValidationError(
          f"Expected a mapping object [{rpath}]: {_obj}")

    _default = default if i == last_i else dict()

    _obj = _obj.get( part, _default )

  return _obj

#===============================================================================
def as_list(obj):
  if isinstance(obj, (str, Mapping)) or not isinstance(obj, Iterable):
    return [obj]

  return list(obj)



from __future__ import annotations
from pathlib import Path
import logging
from .builder import (
  ProcessRunner)
from ..validate import (
  ValidPathError)

#===============================================================================
def process(
  pyproj,
  logger: logging.Logger,
  options: dict,
  work_dir: Path,
  src_dir: Path,
  build_dir: Path,
  prefix: Path,
  setup_args: list[str],
  compile_args: list[str],
  install_args: list[str],
  build_clean: bool,
  runner: ProcessRunner):
  """Run general three-part set of commands
  """

  if not ( build_dir.exists() and any(build_dir.iterdir()) ):
    # build directory is clean
    ...

  elif not build_clean:
    # skip setup if the build directory should be 'clean'
    setup_args = list()

  else:
    raise ValidPathError(
      f"'build_dir' is not empty, remove manually if this is intended or set 'build_clean = false': {build_dir}")


  for cmd in [setup_args, compile_args, install_args]:
    if cmd:
      runner.run(cmd)




from __future__ import annotations
import os
from pathlib import (
  Path)
import tempfile
import shutil
import zipfile
from .dist_base import dist_base
from ..norms import (
  norm_path,
  norm_data,
  norm_zip_external_attr )

#===============================================================================
class dist_zip( dist_base ):
  """Builds a zip file

  Example
  -------

  .. code-block:: python

    import tempfile

    with tempfile.TemporaryDirectory() as tmpdir:

      import os
      import os.path

      pkg_dir = os.path.join( tmpdir, 'src', 'my_package' )
      out_dir = os.path.join( tmpdir, 'build' )

      os.makedirs( pkg_dir )

      with open( os.path.join( pkg_dir, 'module.py' ), 'w' ) as fp:
        fp.write("print('hello')")

      from partis.pyproj import (
        dist_zip )

      with dist_zip(
        outname = 'my_dist.zip',
        outdir = out_dir ) as dist:

        dist.copytree(
          src = pkg_dir,
          dst = 'my_package' )

  """

  #-----------------------------------------------------------------------------
  def __init__( self,
    outname,
    outdir = None,
    tmpdir = None,
    named_dirs = None,
    logger = None ):

    super().__init__(
      outname = outname,
      outdir = outdir,
      tmpdir = tmpdir,
      named_dirs = named_dirs,
      logger = logger )

    self._fd = None
    self._fp = None
    self._tmp_path = None
    self._zipfile = None

  #-----------------------------------------------------------------------------
  def create_distfile( self ):

    ( self._fd, self._tmp_path ) = tempfile.mkstemp(
      dir = self.tmpdir )

    self._tmp_path = Path(self._tmp_path)

    self._fp = os.fdopen( self._fd, "w+b" )

    self._zipfile = zipfile.ZipFile(
      self._fp,
      mode = "w",
      compression = zipfile.ZIP_DEFLATED )

  #-----------------------------------------------------------------------------
  def close_distfile( self ):

    if self._zipfile is not None:

      # close the file
      self._zipfile.close()
      self._zipfile = None

    if self._fp is not None:
      self._fp.close()
      self._fp = None

    if self._fd is not None:
      self._fd = None

  #-----------------------------------------------------------------------------
  def copy_distfile( self ):
    if not self._tmp_path:
      return

    # overwiting in destination directory
    if self.outpath.exists():
      # NOTE: the missing_ok parameter was not added until py38
      self.outpath.unlink()

    self.outdir.mkdir(parents = True, exist_ok = True )
    shutil.copyfile( self._tmp_path, self.outpath )

  #-----------------------------------------------------------------------------
  def remove_distfile( self ):
    if not self._tmp_path:
      return

    # remove temporary file
    if self._tmp_path.exists():
      self._tmp_path.unlink()

    self._tmp_path = None

  #-----------------------------------------------------------------------------
  def write( self,
    dst,
    data,
    mode = None,
    record = True ):

    self.assert_open()

    dst = norm_path( os.fspath(dst) )

    data = norm_data( data )

    zinfo = zipfile.ZipInfo( dst )

    zinfo.external_attr = norm_zip_external_attr( mode )

    self._zipfile.writestr(
      zinfo,
      data,
      compress_type = zipfile.ZIP_DEFLATED )

    super().write(
      dst = dst,
      data = data,
      mode = mode,
      record = record )

  #-----------------------------------------------------------------------------
  def finalize( self ): # pragma: no cover
    pass

  #-----------------------------------------------------------------------------
  def exists( self,
    dst ):

    self.assert_open()

    try:
      self._zipfile.getinfo(os.fspath(dst))
      return True
    except KeyError as e:
      return False



from __future__ import annotations
import os
import io
from copy import copy
import re
import configparser

from pathlib import (
  Path,
  PurePath,
  PurePosixPath)

from .validate import (
  ValidationError,
  validating,
  valid_type,
  valid_keys )

from .norms import (
  norm_path,
  norm_data,
  norm_zip_external_attr,
  hash_sha256,
  email_encode_items )

from .pep import (
  norm_printable,
  valid_dist_name,
  norm_dist_name,
  norm_dist_version,
  norm_dist_author,
  norm_dist_classifier,
  norm_dist_keyword,
  norm_dist_url,
  norm_dist_extra,
  norm_dist_build,
  norm_dist_compat,
  compress_dist_compat,
  norm_dist_filename,
  norm_entry_point_group,
  norm_entry_point_name,
  norm_entry_point_ref )

from .pptoml import project as pptoml_project

import configparser

from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet
from packaging.markers import Marker

#===============================================================================
class EntryPointsParser(configparser.ConfigParser):
  """

  See Also
  --------
  * https://packaging.python.org/en/latest/specifications/entry-points/
  """
  optionxform = staticmethod(str)

#===============================================================================
class PkgInfoAuthor:
  """Internal container for normalizing Author/Maintainer
  and Author-email/Maintainer-email header metadata
  """
  #-----------------------------------------------------------------------------
  def __init__( self, name = '', email = '' ):
    # Note, the normalization will combine "name" <email> into email if both are provided

    # > PEP 621
    # > If only name is provided, the value goes in Author/Maintainer as appropriate.
    # > If only email is provided, the value goes in Author-email/Maintainer-email as appropriate.
    # > If both email and name are provided, the value goes in Author-email/Maintainer-email as
    # > appropriate, with the format {name} <{email}> (with appropriate quoting, e.g. using email.headerregistry.Address).
    self.name, self.email = norm_dist_author(
      name = str(name),
      email = str(email) )

  #-----------------------------------------------------------------------------
  def __str__( self ):
    return self.name + self.email

  #-----------------------------------------------------------------------------
  def __eq__( self, other ):
    return str(self) == str(other)

  #-----------------------------------------------------------------------------
  def __hash__( self ):
    return hash(str(self))

#===============================================================================
class PkgInfoURL:
  """Internal container for normalizing Project-URL
  """
  #-----------------------------------------------------------------------------
  def __init__( self, label = '', url = '' ):
    self.label, self.url = norm_dist_url(
      label = label,
      url = url )

  #-----------------------------------------------------------------------------
  def __str__( self ):
    return f'{self.label}, {self.url}'

  #-----------------------------------------------------------------------------
  def __eq__( self, other ):
    return str(self) == str(other)

  #-----------------------------------------------------------------------------
  def __hash__( self ):
    return hash(str(self))

#===============================================================================
class PkgInfoReq:
  """Internal container for normalizing "Requires-Dist" header metadata
  """
  #-----------------------------------------------------------------------------
  def __init__( self, req, extra = '' ):

    self.req = Requirement( norm_printable(req) )

    marker = str( self.req.marker ) if self.req.marker else ''
    extra = norm_dist_extra(extra)

    if extra:
      if marker:
        self.req.marker = Marker(f'extra == "{extra}" and ( {marker} )')
      else:
        self.req.marker = Marker(f'extra == "{extra}"')

  #-----------------------------------------------------------------------------
  def __str__( self ):
    return str(self.req)

  #-----------------------------------------------------------------------------
  def __eq__( self, other ):
    return str(self) == str(other)

  #-----------------------------------------------------------------------------
  def __hash__( self ):
    return hash(str(self))

#===============================================================================
class PkgInfo:
  def __init__( self,
    project,
    root = None ):
    """Internal container for normalizing metadata as defined in PEP 621 and


    Parameters
    ----------
    project : dict
      The project meta-data as defined in 'pyproject.toml'.
      May be the parsed [project] table from a 'pyproject.toml' file located
      in the 'root' directory.
    root : None | str | pathlib.Path
      Path to the root project directory that would contain 'pyproject.toml'.
      This is used to resolve file paths defined in the project metatada.
      If there are no files referenced, then this value has no effect.

    See Also
    --------
    * https://www.python.org/dev/peps/pep-0621/
    * https://packaging.python.org/en/latest/specifications/core-metadata/
    """

    if not isinstance(project, pptoml_project):
      project = pptoml_project(project)

    root = Path(root) if root else None

    with validating(key = 'dynamic'):
      if project.dynamic:
        raise ValidationError(
          f"All dynamic metadata must be resolved before constructing core metadata: {project.dynamic}")

    self.name = project.name
    self.name_normed = norm_dist_name( self.name )
    self.version = project.version
    self.description = project.description
    self.readme = project.get('readme', None)
    self.license = project.get('license', None)
    self.requires_python = SpecifierSet( project.requires_python )
    self.keywords = set(project.keywords)
    self.classifiers = set(project.classifiers)

    self.dependencies = set([ PkgInfoReq( req = d )
      for d in project.dependencies ])

    self.optional_dependencies = {
      extra : set([
        PkgInfoReq( req = d, extra = extra )
        for d in deps ])
      for extra, deps in project.optional_dependencies.items() }

    self.urls = set([
      PkgInfoURL( label = k, url = v )
      for k,v in project.urls.items() ])

    self.authors = set([ PkgInfoAuthor(**kw)
      for kw in project.authors ])

    self.maintainers = set([ PkgInfoAuthor(**kw)
      for kw in project.maintainers ])

    # NOTE: cannot use the pptoml validated dict
    # since it does not allow combined entry points as in core meta-data
    self.entry_points = dict(project.entry_points)

    # TODO: validate/normalize entrypoints
    if project.scripts:
      self.entry_points['console_scripts'] = project.scripts

    if project.gui_scripts:
      self.entry_points['gui_scripts'] = project.gui_scripts

    #...........................................................................
    # > PEP 621
    # > If the file path ends in a case-insensitive .md suffix, then tools MUST assume
    # > the content-type is text/markdown. If the file path ends in a case-insensitive
    # > .rst, then tools MUST assume the content-type is text/x-rst.
    # > If a tool recognizes more extensions than this PEP, they MAY infer the
    # > content-type for the user without specifying this field as dynamic.
    # > For all unrecognized suffixes when a content-type is not provided, tools MUST
    # > raise an error.
    # TODO: inspect for content-type in file?

    self._long_desc = ''
    self._desc_type = 'text/plain'

    if self.readme:
      with validating(key = 'readme'):
        if 'file' in self.readme:

          if not root:
            raise ValidationError(
              f"'root' must be given to resolve a 'readme.file' path")

          readme_file = root.joinpath(self.readme.file)

          if readme_file.suffix == '.rst':
            self._desc_type = 'text/x-rst'

          elif readme_file.suffix == '.md':
            self._desc_type = 'text/markdown'

          if not readme_file.exists():
            raise ValidationError(
              f"'readme' file not found: {readme_file}")

          with open( readme_file, 'rb' ) as fp:
            self._long_desc = norm_printable(
              fp.read().decode('utf-8', errors = 'replace') )

        else:
          # NOTE: if readme is non-empty, then it must either have 'file' or 'text'
          self._long_desc = self.readme.text

    if not self._long_desc:
      self._long_desc = self.description

    #...........................................................................
    # https://www.python.org/dev/peps/pep-0621/#license
    self._license = ''
    self.license_file = ''
    self.license_file_content = None

    if self.license:
      with validating(key = 'license'):
        # NOTE: PEP 621 specifically says
        # > The text key has a string value which is the license of the project
        # > whose meaning is that of the License field from the core metadata.
        # > These keys are mutually exclusive, so a tool MUST raise an error
        # > if the metadata specifies both keys.
        # However, many tools seem to assign both a 'short' license description
        # to License, in addition to a filename to 'License-File'.
        # It's not clear how to accomidate both with the above restriction.


        # > The table may have one of two keys. The file key has a string value that is
        # > a relative file path to the file which contains the license for the project.
        # > Tools MUST assume the file's encoding is UTF-8. The text key has a string
        # > value which is the license of the project whose meaning is that of the
        # > License field from the core metadata. These keys are mutually exclusive,
        # > so a tool MUST raise an error if the metadata specifies both keys.

        if 'file' in self.license:
          if not root:
            raise ValidationError(f"'root' must be given to resolve 'license.file' path")

          # if 'text' in self.license:
          #   raise ValidationError(f"'license' cannot have both 'text' and 'file': {self.license}")

          # TODO: Core Metadata standar does not mention a 'License-File' header
          # but many tools seem to assign this value.
          # https://packaging.python.org/en/latest/specifications/core-metadata/
          # It is not clear if this is now deprecated, or if any tools actually
          # expect this to be set

          self.license_file = os.fspath(PurePosixPath(self.license.file))

          license_file = root.joinpath(self.license.file)

          if not license_file.exists():
            raise ValidationError(
              f"'license' file not found: {license_file}")

          with open( license_file, 'rb' ) as fp:
            self.license_file_content = norm_printable(
              fp.read().decode('utf-8', errors = 'replace') ).encode('utf-8')

        if 'text' in self.license:
          self._license = norm_printable( self.license.text )

  #-----------------------------------------------------------------------------
  def add_dependencies( self, deps ):
    """Used to add dependencies

    Parameters
    ----------
    deps : List[ str ]
      dependencies to add

    Returns
    -------
    pkg_info :
      Resulting package info

    """

    new_info = copy(self)

    # NOTE: '|' for sets of requirements results in 'require all'
    new_info.dependencies |= set([ PkgInfoReq( req = d )
      for d in deps ])

    return new_info

  #-----------------------------------------------------------------------------
  @property
  def requires_dist( self ):
    """Computes total list of install requirements
    """
    requires_dist = list(self.dependencies)

    for extra, reqs in self.optional_dependencies.items():
      requires_dist.extend( list(reqs) )

    return requires_dist

  #-----------------------------------------------------------------------------
  @property
  def provides_extra( self ):
    """Provided extras
    """
    return list( self.optional_dependencies.keys() )

  #-----------------------------------------------------------------------------
  def encode_entry_points( self ):
    """Generate encoded content for .dist_info/entry_points.txt

    Returns
    -------
    content : bytes
    """

    entry_points = EntryPointsParser()

    for k, v in self.entry_points.items():
      entry_points[k] = v

    fp = io.StringIO()

    entry_points.write( fp )

    return fp.getvalue().encode('utf-8')

  #-----------------------------------------------------------------------------
  def encode_pkg_info( self ):
    """Generate encoded content for PKG-INFO, or .dist_info/METADATA

    Returns
    -------
    content : bytes
    """

    #...........................................................................
    # filter non-empty normalized author fields
    _authors = [ a.name for a in self.authors if a.name ]
    _author_emails = [ a.email for a in self.authors if a.email ]

    _maintainers = [ a.name for a in self.maintainers if a.name ]
    _maintainer_emails = [ a.email for a in self.maintainers if a.email ]


    #...........................................................................
    # construct metadata header values
    headers = [
      ( 'Metadata-Version', '2.1' ),
      ( 'Name', self.name ),
      ( 'Version', self.version ) ]

    if self.requires_python:
      headers.append( ( 'Requires-Python', str(self.requires_python) ) )

    #...........................................................................
    for name in _authors:
      headers.append( ( 'Author', name ) )

    for name in _maintainers:
      headers.append( ( 'Maintainer', name ) )

    for email in _author_emails:
      headers.append( ( 'Author-email', email ) )

    for email in _maintainer_emails:
      headers.append( ( 'Maintainer-email', email ) )

    summary_folded = re.sub(
      r'\n',
      '\n        |',
      self.description.strip() )

    headers.append( ( 'Summary', summary_folded ) )

    if self._license:
      license_folded = re.sub(
        r'\n',
        '\n        |',
        self._license.strip() )

      headers.append( ( 'License', license_folded ) )

    if self.license_file:
      headers.append( ( 'License-File', self.license_file ) )

    if len(self.keywords) > 0:
      headers.append( ( 'Keywords', ', '.join(self.keywords) ) )

    for url in self.urls:
      headers.append( ( 'Project-URL', str(url) ) )

    for classifier in self.classifiers:
      headers.append( ( 'Classifier', classifier ) )

    #...........................................................................
    for e in self.provides_extra:
      headers.append(
        ( 'Provides-Extra', str(e) ) )

    for d in self.requires_dist:
      headers.append(
        ( 'Requires-Dist', str(d) ) )

    headers.append( ( 'Description-Content-Type', self._desc_type ) )

    return email_encode_items(
      headers = headers,
      payload = self._long_desc )



from __future__ import annotations
import sys
import os
import os.path as osp
from io import (
  IOBase,
  BytesIO)
import warnings
import stat
import re
import pathlib
import inspect
from copy import copy
from collections.abc import (
  Mapping,
  Sequence,
  Iterable )

from collections import namedtuple
import hashlib
from base64 import urlsafe_b64encode
from email.message import Message
from email.generator import BytesGenerator
from email.utils import parseaddr, formataddr
from urllib.parse import urlparse
import keyword

from packaging.tags import sys_tags
from packaging.markers import Marker

from .validate import (
  ValidationError,
  valid,
  valid_type,
  valid_keys,
  valid_dict,
  valid_list,
  mapget,
  as_list)

#===============================================================================
def marker_evaluated(marker):
  if isinstance(marker, bool):
    return marker

  marker = Marker(marker)

  return marker.evaluate()

#===============================================================================
def scalar(val):
  if isinstance(val, str):
    return val

  if isinstance(val, Iterable) or isinstance(val, Mapping):
    raise ValidationError(f"Must be a scalar: {val}")

  return val

#===============================================================================
class scalar_list(valid_list):
  value_valid = valid(scalar)

#===============================================================================
def norm_bool(val):
  t = [True, 'true', 'True', 'yes', 'y', 'enable', 'enabled']
  f = [False, 'false', 'False', 'no', 'n', 'disable', 'disabled']

  if val not in t + f:
    raise ValidationError(
      f"Value could not be interpreted as boolean: {val}")

  val = True if val in t else False

  return val

#===============================================================================
def empty_str(val):
  val = str(val)

  if len(val):
    raise ValidationError(
      f"Must be empty string")

  return val

#===============================================================================
def nonempty_str(val):
  val = str(val)

  if len(val) == 0:
    raise ValidationError(
      f"Must be non-empty string: {val}")

  return val

#===============================================================================
class str_list(valid_list):
  value_valid = valid(str)

#===============================================================================
class nonempty_str_list(valid_list):
  value_valid = valid(nonempty_str)

#===============================================================================
def norm_data(data: str|bytes|IOBase) -> bytes:
  """Normalize data for writing into a distribution

  Parameters
  ----------
  data:
    Data to process

  Returns
  -------
  data:

    * If data is bytes, it will be returned un-modified.
    * If data is a str, it will be encoded as 'utf-8'.
    * If data is a stream, it will be read to EOF and apply one of the above.
  """

  if isinstance(data, IOBase):
    # read from a stream, assumes readable
    data = data.read()

  if isinstance(data, str):
    # encode text as utf-8
    data = data.encode('utf-8')

  return data

#===============================================================================
def norm_path( path ):
  """Normalizes a file path for writing into a distribution archive

  Note
  ----
  * Must be a valid path
  * Must be relative, and no reference to parent any directory (e.g. '..')
  * May not contain any white-space in path components
  * All slashes replaced with forward slashes
  """

  path = str(path)

  if re.search( r'\s+', path ):
    raise ValidationError(
      f"path segments should not contain whitespace: {path}")

  # NOTE: starting with assuming windows path leads to the same result wether or
  # not it actually was a windows path, replacing slashes etc as necessary.
  # This should handle whether a path was passed in already posix-like, even when
  # on Windows.
  wpath = pathlib.PureWindowsPath( path )
  path = wpath.as_posix()

  for segment in path.split('/'):
    if windows_invalid_filename.fullmatch( segment ):
      raise ValidationError(
        f"Path segment '{segment}' invalid on Windows platform: {path}")

  ppath = pathlib.PurePosixPath( path )

  if wpath.is_absolute() or ppath.is_absolute():
    raise ValidationError(f"path must be relative: {path}")


  if re.search( r'(\.\.)', path ):
    raise ValidationError(
      f"path segments can not be relative to parent directories: {path}")

  return path

#===============================================================================
def norm_path_to_os( path ):
  """Converts a normalized or OS path to be an OS path
  """

  path = str(path)

  # NOTE: starting with assuming windows path leads to the same result wether or
  # not it actually was a windows path, replacing slashes etc as necessary.
  # This should handle whether a path was passed in already posix like even when
  # on Windows.
  wpath = pathlib.PureWindowsPath(path)
  path = wpath.as_posix()

  return str(pathlib.PurePath(path))

#===============================================================================
def norm_mode( mode = None ):
  """Normalizes file permission mode for distribution archive

  Parameters
  ----------
  mode : None | int
    POSIX permission mode

  Returns
  -------
  int

  Note
  ----
  The returned mode is either ``0o644`` (rw-r--r--), or `0o755` (rwxr-xr-x)
  if ``mode & stat.S_IXUSR == True``

  Example
  -------

  .. code-block:: python

    from partis.pyproj import norm_mode

    print( norm_mode( 0o000 ) == 0o644 )
    print( norm_mode( 0o100 ) == 0o755 )

  """

  if mode is None:
    mode = 0

  mode = int(mode)

  if mode & stat.S_IXUSR:
    # if the mode is executable by the user,
    # set as executable also by group and others

    # rwxr-xr-x
    _mode = 0o755

  else:
    # set as writable by owner, and readable by all others
    # rw-r--r--
    _mode = 0o644

  return _mode

#===============================================================================
def norm_zip_external_attr( mode = None ):
  """Converts the unix integer mode to zip external_attr

  The returned value follows the 4 byte format
  ``|mm|mm|xx|dx|``

  The file permission mode is masked and shifted to the two most
  significant bytes.
  If specified as a directory, the second bit is set.

  Parameters
  ----------
  mode : int

  Returns
  -------
  external_attr : int
  """


  xattr = norm_mode( mode )

  # mask and shift to 3rd byte
  xattr &= 0xFFFF
  xattr = xattr << 16

  # set directory flag after shift
  # MS-DOS directory flag 0x10 (various sources,
  # e.g. https://stackoverflow.com/questions/434641 )
  # NOTE: not sure why this is necessary, except maybe for adding an empty directory
  # if stat.S_ISDIR(mode):
  #   xattr |= 0x10

  return xattr

#===============================================================================
def b64_nopad( data ):
  """Encodes hash as urlsafe base64 encoding with no trailing '=' (:pep:`427`)

  Parameters
  ----------
  data : bytes

  Returns
  -------
  str

  See Also
  --------
  * https://www.python.org/dev/peps/pep-0427/#appendix
  """
  return urlsafe_b64encode( data ).decode("ascii").rstrip("=")

#===============================================================================
def hash_sha256(stream: BytesIO):
  """Computes SHA-256 hash

  Parameters
  ----------
  stream:

  Returns
  -------
  str, int
    urlsafe base64 encoded hash, and size (in bytes) of the hashed data

  See Also
  --------
  :func:`hashlib.sha256`
  """

  if isinstance( stream, bytes ):
    with BytesIO( stream ) as _stream:
      return hash_sha256( _stream )

  hasher = hashlib.sha256()
  size = 0
  bufsize = 2**15

  while True:
    _data = stream.read( bufsize )
    _size = len(_data)

    if _size == 0:
      break

    size += _size
    hasher.update( _data )

  digest = hasher.digest()

  digest_b64_nopad = b64_nopad( digest )

  return digest_b64_nopad, size

#===============================================================================
def email_encode_items(
  headers,
  payload = None ):
  """Encodes a list of headers and a payload

  Parameters
  ----------
  headers : List[ Tuple[str, str ] ]
  payload : str

  Returns
  -------
  bytes

  See Also
  --------
  :mod:`email.message`

  """

  msg = Message()

  for k, v in headers:
    msg[k] = v

  if payload is not None:
    msg.set_payload( payload )

  buffer = BytesIO()

  gen = BytesGenerator(
    buffer,
    mangle_from_ = False,
    maxheaderlen = 0 )

  gen.flatten( msg )

  # TODO: why does ``wheel`` replace newlines with only carriage returns?
  # bytes = buffer.getvalue().replace(b'\r\n', b'\r')
  bytes = buffer.getvalue()

  return bytes

#===============================================================================
class TimeEncode:
  digits = '0123456789abcdefghijklmnopqrstuvwxyz'

  def __init__(self,
    resolution = 60,
    rollover = 60*(int('9zzz',36) + 1),
    width = 4,
    base = 36 ):
    """Encode unix timestamp

    Parameters
    ----------
    resolution : int
      Number of seconds that are resolved
    rollover : int
      Number of seconds upper bound before wrapping.
      The number of distinct values is ``rollover // resolution``
    width : int
      Minimal witdth of encoded number
    base : int
      Numeric base used to encode number
    """

    assert base <= len(self.digits)

    self.resolution = resolution
    self.rollover = rollover
    self.modulus = rollover // resolution
    self.width = width
    self.base = base

  #-----------------------------------------------------------------------------
  @property
  def max(self):
    return self.encode( self.resolution * (self.modulus-1) )

  #-----------------------------------------------------------------------------
  def encode(self, num):
    num = ( int(num) // self.resolution ) % self.modulus
    res = []
    base = self.base
    digits = self.digits

    while num:
      num, i = divmod(num, base)
      res.append(digits[i])

    ts = ''.join(reversed(res))

    return f'{ts:0>{self.width}}'

#===============================================================================
# File/directory names that are invalid on windows
windows_invalid_filename = '|'.join([
  # filename ending with a period
  r'(.+\.)',
  # any characters in the range [0-31], or special characters <, >, :, \, ", ?, *
  r'((?=.*[\x00-\x1f<>:\"?*]).+)',
  # Any of these names, or followed immediately by an extension;
  # for example, NUL.txt
  *[ rf'({name}(\.\w+)?)'
    for name in [
      'CON', 'PRN', 'AUX', 'NUL', 'COM1', 'COM2', 'COM3', 'COM4', 'COM5', 'COM6',
      'COM7', 'COM8', 'COM9', 'LPT1', 'LPT2', 'LPT3', 'LPT4', 'LPT5', 'LPT6',
      'LPT7', 'LPT8', 'LPT9' ] ] ])

windows_invalid_filename = re.compile(
  rf"^({windows_invalid_filename})$",
  re.VERBOSE | re.IGNORECASE | re.UNICODE )



from __future__ import annotations
import sys
import os
import os.path as osp
import io
import warnings
import stat
import re
import pathlib
import inspect
from copy import copy
from collections.abc import (
  Mapping,
  Sequence,
  Iterable )

from collections import namedtuple
import hashlib
from base64 import urlsafe_b64encode
from email.message import Message
from email.generator import BytesGenerator
from email.utils import parseaddr, formataddr
from urllib.parse import urlparse
import keyword

from packaging.tags import sys_tags
from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet

from .validate import (
  ValidationError,
  validating,
  valid_type,
  valid_keys,
  valid_dict,
  valid_list,
  mapget,
  as_list)

#===============================================================================
CompatibilityTags = namedtuple('CompatibilityTags', ['py_tag', 'abi_tag', 'plat_tag'])

#===============================================================================
# NOTE: patterns used for validation are defined at the end of this file

#===============================================================================
class PEPValidationError( ValidationError ):
  """Error from value incompatible with a :term:`PEP`

  Parameters
  ----------
  pep : int
    The referenced PEP number
  msg : str
    Error message
  val : object
    Value that was being validated
  """

  def __init__( self, *, pep, msg, val ):

    msg = inspect.cleandoc( msg )

    super().__init__(
      msg = f'{msg} (PEP {pep}): {val}' )

#===============================================================================
def norm_printable(
  text = None ):
  r"""Removes leading and trailing whitespace and all non-printable characters,
  except for newlines '\\n' and tabs '\\t'.

  Parameters
  ----------
  text : None | str
    If None, an empty string is returned.

  Returns
  -------
  str

  Note
  ----
  While not explicitly stated in any PEP, it is implied through referenced RFCs
  and other assumptions that text in package meta-data should only contain
  printable unicode characters.

  Example
  -------

  .. code-block:: python

    import re
    from partis.pyproj import norm_printable

    x = ''.join([ chr(i) for i in range(50) ])
    print( x.isprintable() )

    y = norm_printable(x)
    print( y.isprintable() )

    z = re.sub(r'[\t\n]', '', y)
    print( z.isprintable() )

    print( norm_printable(None) )
    print( norm_printable('f\ubaaar') )

  """

  if text is None:
    return ''


  return nonprintable.sub( '', str(text).strip() )

#===============================================================================
def valid_dist_name( name ):
  """Checks for valid distribution name (:pep:`426`)

  See Also
  --------
  * https://www.python.org/dev/peps/pep-0426/#name
  """

  name = norm_printable( name )

  if not pep426_dist_name.fullmatch( name ):
    raise PEPValidationError(
      pep = 426,
      msg = "Distribution names MUST ASCII letters, digits, _, -, ., start and end with an ASCII letter or digit",
      val = name )

  return name

#===============================================================================
def norm_dist_name( name ):
  """Normalizes a distribution name (:pep:`503`)

  Note
  ----
  The name should be lowercased with all runs of the
  characters ., -, or _ replaced with a single - character.

  See Also
  --------
  * :func:`valid_dist_name`
  * https://www.python.org/dev/peps/pep-0503/#normalized-names
  """

  name = valid_dist_name( name ).lower()

  # > The name should be lowercased with all runs of the
  # > characters ., -, or _ replaced with a single - character.
  name = re.sub( r'[\-\_\.]+', "-", name )

  return name

#===============================================================================
def norm_dist_filename( name ):
  """Normalize distribution filename component (:pep:`427`)

  Note
  ----
  Each component of the filename is escaped by replacing runs of
  non-alphanumeric characters with an underscore '_'

  See Also
  --------
  * https://www.python.org/dev/peps/pep-0427/#file-name-convention
  """

  return re.sub( r"[^\w\d\.]+", "_", name )

#===============================================================================
def join_dist_filename( parts ):
  """Joins distribution filename component (:pep:`427`)

  Note
  ----
  Each component of the filename is joined by '-'

  See Also
  --------
  * :func:`norm_dist_filename`
  * https://www.python.org/dev/peps/pep-0427/#file-name-convention
  """

  return '-'.join([
    norm_dist_filename(p)
    for p in parts
    if p != ''])

#===============================================================================
def norm_dist_version( version ):
  """Checks for valid distribution version (:pep:`440`)

  See Also
  --------
  * https://www.python.org/dev/peps/pep-0440/#version-scheme
  """

  version = norm_printable( version )

  if not pep440_version.fullmatch( version ):
    raise PEPValidationError(
      pep = 440,
      msg = """Public version identifiers MUST comply with the following scheme,
        [N!]N(.N)*[{a|b|rc}N][.postN][.devN]""",
      val = version )

  return version

#===============================================================================
def norm_dist_author(
  name = None,
  email = None ):
  """Checks for valid distribution author/maintainer name/email (:pep:`621`)

  * The name value MUST be a valid email name
    (i.e. whatever can be put as a name, before an email, in RFC #822)
    and not contain commas.
  * If only name is provided, the value goes in Author/Maintainer as
    appropriate.
  * If only email is provided, the value goes in Author-email/Maintainer-email
    as appropriate.
  * If both email and name are provided, the value goes in
    Author-email/Maintainer-email as appropriate,
    with the format {name} <{email}> (with appropriate quoting,
    e.g. using email.headerregistry.Address).

    .. note::

      The returned name field will be empty in this case.


  Parameters
  ----------
  name : str
  email : str

  Returns
  -------
  name : str
  email : str

  See Also
  --------
  * https://www.python.org/dev/peps/pep-0621/#authors-maintainers
  """

  val = norm_dist_author_dict(dict(name = name, email = email))

  #.............................................................................
  # > If both email and name are provided, the value goes in
  # > Author-email/Maintainer-email as appropriate, with the
  # > format {name} <{email}>.
  if name and email:
    return '', formataddr( (name, email) )

  # > If only name is provided, the value goes in Author/Maintainer as
  # > appropriate.
  # > If only email is provided, the value goes in Author-email/Maintainer-email
  # > as appropriate.
  return name, email

#===============================================================================
def norm_dist_author_dict(val):

  name = norm_printable( val.get('name', '') )
  email = norm_printable( val.get('email', '') )

  _name = name or "Placeholder Name"
  _email = email or "place@holder.com"

  # ensure that at least that the standard Python library can understand the
  # "name" <email> combination
  _name, _email = parseaddr( formataddr( (_name, _email) ) )

  #.............................................................................
  with validating(key = 'name'):
    if name and _name != name:
      raise PEPValidationError(
        pep = 621,
        msg = "The name value MUST be a valid email name, and not contain commas",
        val = name )

    if not pep621_author_name.fullmatch(name):
      raise PEPValidationError(
        pep = 621,
        msg = "The name value MUST be a valid email name, and not contain commas",
        val = name )

  #.............................................................................
  with validating(key = 'email'):
    if email and _email != email:
      raise PEPValidationError(
        pep = 621,
        msg = "The email value MUST be a valid email address",
        val = email )

    if not pep621_author_email.fullmatch(email):
      raise PEPValidationError(
        pep = 621,
        msg = "The email value MUST be a valid email address",
        val = email )

  val = {
    'name': name,
    'email': email }

  return val

#===============================================================================
def norm_dist_classifier( classifier ):
  """
  See Also
  --------
  * https://www.python.org/dev/peps/pep-0301/#distutils-trove-classification
  """

  classifier = norm_printable( classifier )

  parts = [ s.strip() for s in classifier.split('::') ]

  for part in parts:
    if not pep_301_classifier.fullmatch( part ):
      raise PEPValidationError(
        pep = 301,
        msg = f"Invalid classifier component '{part}'",
        val = classifier )

  classifier = ' :: '.join( parts )

  return classifier

#===============================================================================
def norm_dist_keyword( keyword ):
  """
  See Also
  --------
  * https://www.python.org/dev/peps/pep-0621/#keywords
  """

  keyword = norm_printable( keyword )

  if not pep_621_keyword.fullmatch( keyword ):
    raise PEPValidationError(
      pep = 621,
      msg = "Invalid keyword",
      val = keyword )

  return keyword

#===============================================================================
def norm_dist_url( label, url ):
  """
  See Also
  --------
  * https://packaging.python.org/en/latest/specifications/core-metadata/#project-url-multiple-use
  """

  # > The label is free text limited to 32 characters.
  label = norm_printable( label )[:32]
  url = norm_printable( url )

  if not pep621_author_name.fullmatch(label):
    raise PEPValidationError(
      pep = 621,
      msg = "Invalid url label",
      val = label )

  try:
    res = urlparse( url )

    if not ( res.scheme and res.netloc ):
      raise PEPValidationError(
        pep = 621,
        msg = "URL must have a valid scheme and net location",
        val = url )

  except Exception as e:
    raise PEPValidationError(
      pep = 621,
      msg = "Invalid url",
      val = url ) from e

  return label, url

#===============================================================================
def norm_dist_extra( extra ):
  """Normalize distribution 'extra' requirement

  Note
  ----
  * No known PEP specifies this format, but is treated as
  """

  extra = norm_printable( extra )

  if not pep_621_extra.fullmatch( extra ):
    raise PEPValidationError(
      pep = 621,
      msg = "Invalid extra",
      val = extra )

  return extra

#===============================================================================
def dist_build( build_number = None, build_tag = None ):
  if build_number is None and build_tag is None:
    build = ''

  elif build_tag is None:
    build = str(int(build_number))

  elif build_number is None:
    build = f"0_{build_tag}"

  else:
    build = f"{int(build_number)}_{build_tag}"

  return norm_dist_build(build)

#===============================================================================
def norm_dist_build( build ):
  """
  Note
  ----
  * Must start with a digit, remainder is ASCII alpha-numeric

  See Also
  --------
  * https://www.python.org/dev/peps/pep-0427/#file-name-convention
  """

  build = norm_printable( build ).lower()

  if not pep427_build.fullmatch( build ):
    raise PEPValidationError(
      pep = 427,
      msg = """Must start with a digit. Acts as a tie-breaker if two wheel file
        names are the same in all other respects""",
      val = build )

  return build

#===============================================================================
def norm_dist_compat( py_tag, abi_tag, plat_tag ):
  """

  Note
  ----
  * Tags must contain only ASCII alpha-numeric or underscore
  * platform tag with all hyphens -
    and periods . replaced with underscore _.

  See Also
  --------
  * https://www.python.org/dev/peps/pep-0425/#details
  """

  py_tag = norm_printable( py_tag ).lower()
  abi_tag = norm_printable( abi_tag ).lower()

  # > platform tag is simply distutils.util.get_platform() with all hyphens -
  # > and periods . replaced with underscore _.
  plat_tag = re.sub( r'[\-\_\.]+', "_", norm_printable(plat_tag).lower() )

  if not pep425_pytag.fullmatch( py_tag ):
    raise PEPValidationError(
      pep = 425,
      msg = """The version is py_version_nodot. CPython gets away with no dot,
        but if one is needed the underscore _ is used instead""",
      val = py_tag )


  if not pep425_pytag.fullmatch( abi_tag ):
    # use the same validation for abi tag
    raise PEPValidationError(
      pep = 425,
      msg = """The version is py_version_nodot. CPython gets away with no dot,
        but if one is needed the underscore _ is used instead""",
      val = abi_tag )

  if not pep425_pytag.fullmatch( plat_tag ):
    # use the same validation for platform tag
    raise PEPValidationError(
      pep = 425,
      msg = """Platform tag is simply distutils.util.get_platform() with all
        hyphens - and periods . replaced with underscore _""",
      val = plat_tag )

  # if not common_pytag.fullmatch( py_tag ):
  #   warnings.warn(f"python tag was not recognized: {py_tag}")
  #
  # if not common_abitag.fullmatch( abi_tag ):
  #   warnings.warn(f"abi tag was not recognized: {abi_tag}")
  #
  # if not any( plat.fullmatch( plat_tag ) for plat in common_plattag.values() ):
  #   warnings.warn(f"platform tag was not recognized: {plat_tag}")

  return CompatibilityTags( py_tag, abi_tag, plat_tag )

#===============================================================================
def join_dist_compat( tags ):
  """
  See Also
  --------
  * https://www.python.org/dev/peps/pep-0425/#compressed-tag-sets
  """
  return '.'.join( sorted(list(set(tags))) )

#===============================================================================
def compress_dist_compat( compat ):
  """
  See Also
  --------
  * https://www.python.org/dev/peps/pep-0425/#compressed-tag-sets
  """

  py_tags, abi_tags, plat_tags = zip( *compat )

  py_tags = join_dist_compat( py_tags )
  abi_tags = join_dist_compat( abi_tags )
  plat_tags = join_dist_compat( plat_tags )

  return py_tags, abi_tags, plat_tags

#===============================================================================
def purelib_compat_tags():
  """Return general compatability tags for the current system
  """

  compat = [ CompatibilityTags( 'py3', 'none', 'any' ) ]

  return compat

#===============================================================================
def platlib_compat_tags():
  """Get platform compatability tags for the current system
  """
  tag = next(iter(sys_tags()))

  # interpreter = "py{0}{1}".format(sys.version_info.major, sys.version_info.minor)
  interpreter = tag.interpreter

  compat_tags = [ CompatibilityTags( interpreter, tag.abi, tag.platform ) ]

  return compat_tags

#===============================================================================
def norm_py_identifier( name ):

  name = norm_printable( name )

  if not py_identifier.fullmatch( name ):
    raise ValidationError(
      msg = f"""Python identifier may only contain letters in a small case (a-z),
        upper case (A-Z), digits (0-9), and underscore (_), and not start with
        a digit: {name}""" )

  if py_keyword.fullmatch( name ):
    raise ValidationError(
      msg = f"Python identifier may not be a reserved keyword: {name}" )

  return name

#===============================================================================
def norm_entry_point_group( name ):
  """Normalizes entry point group

  See Also
  --------
  * https://packaging.python.org/en/latest/specifications/entry-points/
  """

  name = norm_printable( name )

  if not entry_point_group.fullmatch( name ):
    raise ValidationError(
      msg = f"""Entry point group must be one or more groups of
        letters, numbers and underscores, separated by dots: {name}""" )

  return name

#===============================================================================
def norm_entry_point_name( name ):
  """Normalizes entry point name

  See Also
  --------
  * https://packaging.python.org/en/latest/specifications/entry-points/
  """

  name = norm_printable( name )

  if not entry_point_name.fullmatch( name ):
    raise ValidationError(
      msg = f"""Entry point name must be only letters, numbers, underscores,
        dots and dashes: {name}""" )

  return name

#===============================================================================
def norm_entry_point_ref( ref ):
  """Normalizes entry point object reference

  See Also
  --------
  * https://packaging.python.org/en/latest/specifications/entry-points/
  """

  ref = norm_printable( ref )

  modname, sep, qualname = ref.partition(':')

  if not modname:
    raise ValidationError(
      msg = f"Entry point reference must give a module name: {ref}" )

  try:

    modname = '.'.join( norm_py_identifier(name) for name in modname.split('.') )

    if qualname:
      qualname = '.'.join( norm_py_identifier(name) for name in qualname.split('.') )

      return f'{modname}:{qualname}'

    return modname

  except ValidationError as e:
    raise ValidationError(
      msg = f"""Entry point reference must have the form 'importable.module'
        or 'importable.module:object.attr': {ref}""") from e

#===============================================================================
pep426_dist_name = re.compile(
  r'^([A-Z0-9]|[A-Z0-9][A-Z0-9._\-]*[A-Z0-9])$',
  re.IGNORECASE )

pep440_version = re.compile(
  r'^([1-9][0-9]*!)?(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))*'
  r'((a|b|rc)(0|[1-9][0-9]*))?'
  r'(\.post(0|[1-9][0-9]*))?(\.dev(0|[1-9][0-9]*))?$' )

# NOTE: PEP 427 does not specify any constraints on the string following the
# digits, but given the form it is used in the filenames it really cannot
# contain anything other than alpha-numeric characters.
pep427_build = re.compile(
  r'^([0-9]+[A-Z0-9_]*)?$',
  re.IGNORECASE )

pep425_pytag = re.compile(
  r'^([A-Z0-9_]+)$',
  re.IGNORECASE )

#===============================================================================
# https://www.python.org/dev/peps/pep-0621/#authors-maintainers
# https://www.rfc-editor.org/rfc/inline-errata/rfc5322.html
# > name value MUST be a valid email name (i.e. whatever can be put as a name,
# > before an email, in RFC #822) and not contain commas

# NOTE: email names are notoriously hard to validate correctly,
# this is probably not correct.
# strategy here is to do minimal sanity check by ensuring the absense of
# likely invalid characters.
# The strings should also be checked if they are 'printable'.

# ensures the name does not include double-quotes, backslashes, or linefeeds
# or tabs.
# the '@' here is also included here
pep621_author_name = re.compile( r'^([^\"\\\,\@\r\n\t\f\v]+)?$', re.UNICODE )

# ensures that there is a single `@` separating two non-empty segments,
# and each segment does not contain white-space or another `@`
pep621_author_email = re.compile( r'^([^\@\s]+@[^\@\s]+)?$', re.UNICODE )

local_plat = re.sub(r'[\-\.]', '_', sys.platform )

common_pytag = re.compile( r'^(py|cp|ip|pp|jy)(\w+)$', re.IGNORECASE )
common_abitag = re.compile( r'^(none|cp|abi)(\w*)$', re.IGNORECASE )
common_plattag = {
  'any' : re.compile( r'^(any)$', re.IGNORECASE ),
  'win' : re.compile( r'^(win(32|64))$', re.IGNORECASE ),
  'mac' : re.compile( r'^((macos(x)?|darwin)(_\w+)?)$', re.IGNORECASE ),
  # https://www.python.org/dev/peps/pep-0600/
  # manylinux_${GLIBCMAJOR}_${GLIBCMINOR}_${ARCH}
  'linux' : re.compile(
    r'^((many)?linux(_(\d\d?)_(\d\d?))?'
    r'_(i386|x86_64|i686|aarch64|armv7l|ppc64|ppc64le|s390x))$',
    re.IGNORECASE ),
  'local' : re.compile( rf'^({local_plat})$', re.IGNORECASE )}

#===============================================================================
# https://www.python.org/dev/peps/pep-0301/#distutils-trove-classification
# > It was decided that strings would be used for the classification entries
# > due to the deep nesting that would be involved in a more formal Python
# > structure.
# > ... classification namespaces be separated by ...  double-colon solution
# > (" :: ")
# NOTE: the PEP does not specify a valid form for classifiers, other than
# "The list of classifiers will be available through the web".
# This allows any package or version text, brackets, parentheses, spaces,
# and forward slash.
# TODO: write test against current list of classifiers
pep_301_classifier = re.compile(
  r'^[A-Z0-9._\-\/\[\]\(\) ]+$',
  re.IGNORECASE )

#===============================================================================
# https://packaging.python.org/en/latest/specifications/core-metadata/#keywords
# https://www.python.org/dev/peps/pep-0621/#keywords
# NOTE: does not say what is a valid keyword, but does say they are comma separted,
# and other implemented with space separated. To be safe ensure no white-space or commas
pep_621_keyword = re.compile( r'^[^\s\,]+$' )

#===============================================================================
pep_621_extra = re.compile( r'^([A-Z0-9_]+)?$', re.IGNORECASE  )

#===============================================================================
# https://packaging.python.org/en/latest/specifications/entry-points/
# Group names must be one or more groups of letters, numbers and underscores,
# separated by dots
entry_point_group = re.compile( r'^[A-Z0-9_]+(\.[A-Z0-9_]+)*$', re.IGNORECASE  )
# For new entry points (names), it is recommended to use only letters, numbers,
# underscores, dots and dashes
entry_point_name = re.compile( r'^([A-Z0-9_\.\-]+)?$', re.IGNORECASE  )

#===============================================================================
py_keyword = re.compile( '^(' + '|'.join(keyword.kwlist) + ')$' )
py_identifier = re.compile( r'^[A-Z_][A-Z0-9_]*$', re.IGNORECASE )

#===============================================================================
# NOTE: there may be a more efficient way to strip all non-printable characters
# Here consider new-lines '\n' and tabs '\t' to be printable
# even though '\n'.isprintable() returns False
# see _nonprintable.py for how this was generated
nonprintable = (
  r'[\x00-\x08\x0B-\x1F\x7F-\xA0\xAD\u0378-\u0379\u0380-\u0383\u038B\u038D'
  r'\u03A2\u0530\u0557-\u0558\u058B-\u058C\u0590\u05C8-\u05CF\u05EB-\u05EE'
  r'\u05F5-\u0605\u061C-\u061D\u06DD\u070E-\u070F\u074B-\u074C\u07B2-\u07BF'
  r'\u07FB-\u07FC\u082E-\u082F\u083F\u085C-\u085D\u085F\u086B-\u089F\u08B5'
  r'\u08C8-\u08D2\u08E2\u0984\u098D-\u098E\u0991-\u0992\u09A9\u09B1'
  r'\u09B3-\u09B5\u09BA-\u09BB\u09C5-\u09C6\u09C9-\u09CA\u09CF-\u09D6'
  r'\u09D8-\u09DB\u09DE\u09E4-\u09E5\u09FF-\u0A00\u0A04\u0A0B-\u0A0E'
  r'\u0A11-\u0A12\u0A29\u0A31\u0A34\u0A37\u0A3A-\u0A3B\u0A3D\u0A43-\u0A46'
  r'\u0A49-\u0A4A\u0A4E-\u0A50\u0A52-\u0A58\u0A5D\u0A5F-\u0A65\u0A77-\u0A80'
  r'\u0A84\u0A8E\u0A92\u0AA9\u0AB1\u0AB4\u0ABA-\u0ABB\u0AC6\u0ACA\u0ACE-\u0ACF'
  r'\u0AD1-\u0ADF\u0AE4-\u0AE5\u0AF2-\u0AF8\u0B00\u0B04\u0B0D-\u0B0E'
  r'\u0B11-\u0B12\u0B29\u0B31\u0B34\u0B3A-\u0B3B\u0B45-\u0B46\u0B49-\u0B4A'
  r'\u0B4E-\u0B54\u0B58-\u0B5B\u0B5E\u0B64-\u0B65\u0B78-\u0B81\u0B84'
  r'\u0B8B-\u0B8D\u0B91\u0B96-\u0B98\u0B9B\u0B9D\u0BA0-\u0BA2\u0BA5-\u0BA7'
  r'\u0BAB-\u0BAD\u0BBA-\u0BBD\u0BC3-\u0BC5\u0BC9\u0BCE-\u0BCF\u0BD1-\u0BD6'
  r'\u0BD8-\u0BE5\u0BFB-\u0BFF\u0C0D\u0C11\u0C29\u0C3A-\u0C3C\u0C45\u0C49'
  r'\u0C4E-\u0C54\u0C57\u0C5B-\u0C5F\u0C64-\u0C65\u0C70-\u0C76\u0C8D\u0C91'
  r'\u0CA9\u0CB4\u0CBA-\u0CBB\u0CC5\u0CC9\u0CCE-\u0CD4\u0CD7-\u0CDD\u0CDF'
  r'\u0CE4-\u0CE5\u0CF0\u0CF3-\u0CFF\u0D0D\u0D11\u0D45\u0D49\u0D50-\u0D53'
  r'\u0D64-\u0D65\u0D80\u0D84\u0D97-\u0D99\u0DB2\u0DBC\u0DBE-\u0DBF'
  r'\u0DC7-\u0DC9\u0DCB-\u0DCE\u0DD5\u0DD7\u0DE0-\u0DE5\u0DF0-\u0DF1'
  r'\u0DF5-\u0E00\u0E3B-\u0E3E\u0E5C-\u0E80\u0E83\u0E85\u0E8B\u0EA4\u0EA6'
  r'\u0EBE-\u0EBF\u0EC5\u0EC7\u0ECE-\u0ECF\u0EDA-\u0EDB\u0EE0-\u0EFF\u0F48'
  r'\u0F6D-\u0F70\u0F98\u0FBD\u0FCD\u0FDB-\u0FFF\u10C6\u10C8-\u10CC'
  r'\u10CE-\u10CF\u1249\u124E-\u124F\u1257\u1259\u125E-\u125F\u1289'
  r'\u128E-\u128F\u12B1\u12B6-\u12B7\u12BF\u12C1\u12C6-\u12C7\u12D7\u1311'
  r'\u1316-\u1317\u135B-\u135C\u137D-\u137F\u139A-\u139F\u13F6-\u13F7'
  r'\u13FE-\u13FF\u1680\u169D-\u169F\u16F9-\u16FF\u170D\u1715-\u171F'
  r'\u1737-\u173F\u1754-\u175F\u176D\u1771\u1774-\u177F\u17DE-\u17DF'
  r'\u17EA-\u17EF\u17FA-\u17FF\u180E-\u180F\u181A-\u181F\u1879-\u187F'
  r'\u18AB-\u18AF\u18F6-\u18FF\u191F\u192C-\u192F\u193C-\u193F\u1941-\u1943'
  r'\u196E-\u196F\u1975-\u197F\u19AC-\u19AF\u19CA-\u19CF\u19DB-\u19DD'
  r'\u1A1C-\u1A1D\u1A5F\u1A7D-\u1A7E\u1A8A-\u1A8F\u1A9A-\u1A9F\u1AAE-\u1AAF'
  r'\u1AC1-\u1AFF\u1B4C-\u1B4F\u1B7D-\u1B7F\u1BF4-\u1BFB\u1C38-\u1C3A'
  r'\u1C4A-\u1C4C\u1C89-\u1C8F\u1CBB-\u1CBC\u1CC8-\u1CCF\u1CFB-\u1CFF\u1DFA'
  r'\u1F16-\u1F17\u1F1E-\u1F1F\u1F46-\u1F47\u1F4E-\u1F4F\u1F58\u1F5A\u1F5C'
  r'\u1F5E\u1F7E-\u1F7F\u1FB5\u1FC5\u1FD4-\u1FD5\u1FDC\u1FF0-\u1FF1\u1FF5'
  r'\u1FFF-\u200F\u2028-\u202F\u205F-\u206F\u2072-\u2073\u208F\u209D-\u209F'
  r'\u20C0-\u20CF\u20F1-\u20FF\u218C-\u218F\u2427-\u243F\u244B-\u245F'
  r'\u2B74-\u2B75\u2B96\u2C2F\u2C5F\u2CF4-\u2CF8\u2D26\u2D28-\u2D2C'
  r'\u2D2E-\u2D2F\u2D68-\u2D6E\u2D71-\u2D7E\u2D97-\u2D9F\u2DA7\u2DAF\u2DB7'
  r'\u2DBF\u2DC7\u2DCF\u2DD7\u2DDF\u2E53-\u2E7F\u2E9A\u2EF4-\u2EFF\u2FD6-\u2FEF'
  r'\u2FFC-\u3000\u3040\u3097-\u3098\u3100-\u3104\u3130\u318F\u31E4-\u31EF'
  r'\u321F\u9FFD-\u9FFF\uA48D-\uA48F\uA4C7-\uA4CF\uA62C-\uA63F\uA6F8-\uA6FF'
  r'\uA7C0-\uA7C1\uA7CB-\uA7F4\uA82D-\uA82F\uA83A-\uA83F\uA878-\uA87F'
  r'\uA8C6-\uA8CD\uA8DA-\uA8DF\uA954-\uA95E\uA97D-\uA97F\uA9CE\uA9DA-\uA9DD'
  r'\uA9FF\uAA37-\uAA3F\uAA4E-\uAA4F\uAA5A-\uAA5B\uAAC3-\uAADA\uAAF7-\uAB00'
  r'\uAB07-\uAB08\uAB0F-\uAB10\uAB17-\uAB1F\uAB27\uAB2F\uAB6C-\uAB6F'
  r'\uABEE-\uABEF\uABFA-\uABFF\uD7A4-\uD7AF\uD7C7-\uD7CA\uD7FC-\uF8FF'
  r'\uFA6E-\uFA6F\uFADA-\uFAFF\uFB07-\uFB12\uFB18-\uFB1C\uFB37\uFB3D\uFB3F'
  r'\uFB42\uFB45\uFBC2-\uFBD2\uFD40-\uFD4F\uFD90-\uFD91\uFDC8-\uFDEF'
  r'\uFDFE-\uFDFF\uFE1A-\uFE1F\uFE53\uFE67\uFE6C-\uFE6F\uFE75\uFEFD-\uFF00'
  r'\uFFBF-\uFFC1\uFFC8-\uFFC9\uFFD0-\uFFD1\uFFD8-\uFFD9\uFFDD-\uFFDF\uFFE7'
  r'\uFFEF-\uFFFB\uFFFE-\uFFFF\U0001000C\U00010027\U0001003B\U0001003E'
  r'\U0001004E-\U0001004F\U0001005E-\U0001007F\U000100FB-\U000100FF'
  r'\U00010103-\U00010106\U00010134-\U00010136\U0001018F\U0001019D-\U0001019F'
  r'\U000101A1-\U000101CF\U000101FE-\U0001027F\U0001029D-\U0001029F'
  r'\U000102D1-\U000102DF\U000102FC-\U000102FF\U00010324-\U0001032C'
  r'\U0001034B-\U0001034F\U0001037B-\U0001037F\U0001039E\U000103C4-\U000103C7'
  r'\U000103D6-\U000103FF\U0001049E-\U0001049F\U000104AA-\U000104AF'
  r'\U000104D4-\U000104D7\U000104FC-\U000104FF\U00010528-\U0001052F'
  r'\U00010564-\U0001056E\U00010570-\U000105FF\U00010737-\U0001073F'
  r'\U00010756-\U0001075F\U00010768-\U000107FF\U00010806-\U00010807\U00010809'
  r'\U00010836\U00010839-\U0001083B\U0001083D-\U0001083E\U00010856'
  r'\U0001089F-\U000108A6\U000108B0-\U000108DF\U000108F3\U000108F6-\U000108FA'
  r'\U0001091C-\U0001091E\U0001093A-\U0001093E\U00010940-\U0001097F'
  r'\U000109B8-\U000109BB\U000109D0-\U000109D1\U00010A04\U00010A07-\U00010A0B'
  r'\U00010A14\U00010A18\U00010A36-\U00010A37\U00010A3B-\U00010A3E'
  r'\U00010A49-\U00010A4F\U00010A59-\U00010A5F\U00010AA0-\U00010ABF'
  r'\U00010AE7-\U00010AEA\U00010AF7-\U00010AFF\U00010B36-\U00010B38'
  r'\U00010B56-\U00010B57\U00010B73-\U00010B77\U00010B92-\U00010B98'
  r'\U00010B9D-\U00010BA8\U00010BB0-\U00010BFF\U00010C49-\U00010C7F'
  r'\U00010CB3-\U00010CBF\U00010CF3-\U00010CF9\U00010D28-\U00010D2F'
  r'\U00010D3A-\U00010E5F\U00010E7F\U00010EAA\U00010EAE-\U00010EAF'
  r'\U00010EB2-\U00010EFF\U00010F28-\U00010F2F\U00010F5A-\U00010FAF'
  r'\U00010FCC-\U00010FDF\U00010FF7-\U00010FFF\U0001104E-\U00011051'
  r'\U00011070-\U0001107E\U000110BD\U000110C2-\U000110CF\U000110E9-\U000110EF'
  r'\U000110FA-\U000110FF\U00011135\U00011148-\U0001114F\U00011177-\U0001117F'
  r'\U000111E0\U000111F5-\U000111FF\U00011212\U0001123F-\U0001127F\U00011287'
  r'\U00011289\U0001128E\U0001129E\U000112AA-\U000112AF\U000112EB-\U000112EF'
  r'\U000112FA-\U000112FF\U00011304\U0001130D-\U0001130E\U00011311-\U00011312'
  r'\U00011329\U00011331\U00011334\U0001133A\U00011345-\U00011346'
  r'\U00011349-\U0001134A\U0001134E-\U0001134F\U00011351-\U00011356'
  r'\U00011358-\U0001135C\U00011364-\U00011365\U0001136D-\U0001136F'
  r'\U00011375-\U000113FF\U0001145C\U00011462-\U0001147F\U000114C8-\U000114CF'
  r'\U000114DA-\U0001157F\U000115B6-\U000115B7\U000115DE-\U000115FF'
  r'\U00011645-\U0001164F\U0001165A-\U0001165F\U0001166D-\U0001167F'
  r'\U000116B9-\U000116BF\U000116CA-\U000116FF\U0001171B-\U0001171C'
  r'\U0001172C-\U0001172F\U00011740-\U000117FF\U0001183C-\U0001189F'
  r'\U000118F3-\U000118FE\U00011907-\U00011908\U0001190A-\U0001190B\U00011914'
  r'\U00011917\U00011936\U00011939-\U0001193A\U00011947-\U0001194F'
  r'\U0001195A-\U0001199F\U000119A8-\U000119A9\U000119D8-\U000119D9'
  r'\U000119E5-\U000119FF\U00011A48-\U00011A4F\U00011AA3-\U00011ABF'
  r'\U00011AF9-\U00011BFF\U00011C09\U00011C37\U00011C46-\U00011C4F'
  r'\U00011C6D-\U00011C6F\U00011C90-\U00011C91\U00011CA8\U00011CB7-\U00011CFF'
  r'\U00011D07\U00011D0A\U00011D37-\U00011D39\U00011D3B\U00011D3E'
  r'\U00011D48-\U00011D4F\U00011D5A-\U00011D5F\U00011D66\U00011D69\U00011D8F'
  r'\U00011D92\U00011D99-\U00011D9F\U00011DAA-\U00011EDF\U00011EF9-\U00011FAF'
  r'\U00011FB1-\U00011FBF\U00011FF2-\U00011FFE\U0001239A-\U000123FF\U0001246F'
  r'\U00012475-\U0001247F\U00012544-\U00012FFF\U0001342F-\U000143FF'
  r'\U00014647-\U000167FF\U00016A39-\U00016A3F\U00016A5F\U00016A6A-\U00016A6D'
  r'\U00016A70-\U00016ACF\U00016AEE-\U00016AEF\U00016AF6-\U00016AFF'
  r'\U00016B46-\U00016B4F\U00016B5A\U00016B62\U00016B78-\U00016B7C'
  r'\U00016B90-\U00016E3F\U00016E9B-\U00016EFF\U00016F4B-\U00016F4E'
  r'\U00016F88-\U00016F8E\U00016FA0-\U00016FDF\U00016FE5-\U00016FEF'
  r'\U00016FF2-\U00016FFF\U000187F8-\U000187FF\U00018CD6-\U00018CFF'
  r'\U00018D09-\U0001AFFF\U0001B11F-\U0001B14F\U0001B153-\U0001B163'
  r'\U0001B168-\U0001B16F\U0001B2FC-\U0001BBFF\U0001BC6B-\U0001BC6F'
  r'\U0001BC7D-\U0001BC7F\U0001BC89-\U0001BC8F\U0001BC9A-\U0001BC9B'
  r'\U0001BCA0-\U0001CFFF\U0001D0F6-\U0001D0FF\U0001D127-\U0001D128'
  r'\U0001D173-\U0001D17A\U0001D1E9-\U0001D1FF\U0001D246-\U0001D2DF'
  r'\U0001D2F4-\U0001D2FF\U0001D357-\U0001D35F\U0001D379-\U0001D3FF\U0001D455'
  r'\U0001D49D\U0001D4A0-\U0001D4A1\U0001D4A3-\U0001D4A4\U0001D4A7-\U0001D4A8'
  r'\U0001D4AD\U0001D4BA\U0001D4BC\U0001D4C4\U0001D506\U0001D50B-\U0001D50C'
  r'\U0001D515\U0001D51D\U0001D53A\U0001D53F\U0001D545\U0001D547-\U0001D549'
  r'\U0001D551\U0001D6A6-\U0001D6A7\U0001D7CC-\U0001D7CD\U0001DA8C-\U0001DA9A'
  r'\U0001DAA0\U0001DAB0-\U0001DFFF\U0001E007\U0001E019-\U0001E01A\U0001E022'
  r'\U0001E025\U0001E02B-\U0001E0FF\U0001E12D-\U0001E12F\U0001E13E-\U0001E13F'
  r'\U0001E14A-\U0001E14D\U0001E150-\U0001E2BF\U0001E2FA-\U0001E2FE'
  r'\U0001E300-\U0001E7FF\U0001E8C5-\U0001E8C6\U0001E8D7-\U0001E8FF'
  r'\U0001E94C-\U0001E94F\U0001E95A-\U0001E95D\U0001E960-\U0001EC70'
  r'\U0001ECB5-\U0001ED00\U0001ED3E-\U0001EDFF\U0001EE04\U0001EE20\U0001EE23'
  r'\U0001EE25-\U0001EE26\U0001EE28\U0001EE33\U0001EE38\U0001EE3A'
  r'\U0001EE3C-\U0001EE41\U0001EE43-\U0001EE46\U0001EE48\U0001EE4A\U0001EE4C'
  r'\U0001EE50\U0001EE53\U0001EE55-\U0001EE56\U0001EE58\U0001EE5A\U0001EE5C'
  r'\U0001EE5E\U0001EE60\U0001EE63\U0001EE65-\U0001EE66\U0001EE6B\U0001EE73'
  r'\U0001EE78\U0001EE7D\U0001EE7F\U0001EE8A\U0001EE9C-\U0001EEA0\U0001EEA4'
  r'\U0001EEAA\U0001EEBC-\U0001EEEF\U0001EEF2-\U0001EFFF\U0001F02C-\U0001F02F'
  r'\U0001F094-\U0001F09F\U0001F0AF-\U0001F0B0\U0001F0C0\U0001F0D0'
  r'\U0001F0F6-\U0001F0FF\U0001F1AE-\U0001F1E5\U0001F203-\U0001F20F'
  r'\U0001F23C-\U0001F23F\U0001F249-\U0001F24F\U0001F252-\U0001F25F'
  r'\U0001F266-\U0001F2FF\U0001F6D8-\U0001F6DF\U0001F6ED-\U0001F6EF'
  r'\U0001F6FD-\U0001F6FF\U0001F774-\U0001F77F\U0001F7D9-\U0001F7DF'
  r'\U0001F7EC-\U0001F7FF\U0001F80C-\U0001F80F\U0001F848-\U0001F84F'
  r'\U0001F85A-\U0001F85F\U0001F888-\U0001F88F\U0001F8AE-\U0001F8AF'
  r'\U0001F8B2-\U0001F8FF\U0001F979\U0001F9CC\U0001FA54-\U0001FA5F'
  r'\U0001FA6E-\U0001FA6F\U0001FA75-\U0001FA77\U0001FA7B-\U0001FA7F'
  r'\U0001FA87-\U0001FA8F\U0001FAA9-\U0001FAAF\U0001FAB7-\U0001FABF'
  r'\U0001FAC3-\U0001FACF\U0001FAD7-\U0001FAFF\U0001FB93\U0001FBCB-\U0001FBEF'
  r'\U0001FBFA-\U0001FFFF\U0002A6DE-\U0002A6FF\U0002B735-\U0002B73F'
  r'\U0002B81E-\U0002B81F\U0002CEA2-\U0002CEAF\U0002EBE1-\U0002F7FF'
  r'\U0002FA1E-\U0002FFFF\U0003134B-\U000E00FF\U000E01F0-\U0010FFFF]' )

nonprintable = re.compile( nonprintable, re.UNICODE )



from __future__ import annotations
import os
import os.path as osp
import sys
import importlib
from pathlib import (
  Path,
  PurePath,
  PurePosixPath)
import hashlib
from base64 import b16encode

from .norms import (
  norm_path_to_os )

from .validate import (
  ValidationWarning,
  ValidationError,
  FileOutsideRootError,
  valid_dict,
  validating,
  valid,
  restrict,
  mapget )

#===============================================================================
class EntryPointError(ValueError):
  pass

#===============================================================================
def module_name_from_path( path, root ):
  """Generates an importable module name from a file system path

  Parameters
  ----------
  path : pathlib.Path
    Path to the module directory relative to 'root'
  root : pathlib.Path
    Base path from which the module will be imported
  """

  path_parts = path.relative_to(root).parts

  if len(path_parts) == 0:
    raise EntryPointError("Empty module name")

  return ".".join(path_parts)

#===============================================================================
def module_path( mod_name, root ):
  path = root / Path(*mod_name.split('.'))

  if path.exists() and path.is_dir():
    path /= '__init__.py'
  else:
    path = path.with_suffix('.py')

  if not path.exists():
    path = None

  return path

#===============================================================================
def load_module( name, path, root ):

  path = str(path)

  # TODO: handle module names when module is not directly in 'root'.
  hasher = hashlib.sha256()
  hasher.update( path.encode('utf-8') )
  digest = hasher.digest()

  _prefix = 'pyprojentry_' + b16encode(digest).decode('ascii')

  spec = importlib.util.spec_from_file_location(
    name = _prefix + name,
    location = path )

  mod = importlib.util.module_from_spec( spec )
  sys.modules[ spec.name ] = mod

  spec.loader.exec_module( mod )

  return mod

#===============================================================================
def load_entrypoint( entry_point, root ):
  r"""
  Parameters
  ----------
  entry_point : str
    Entry point spec
  root : pathlib.Path
    Root project directory.
  """

  mod_name, attr_name = entry_point.split(':')

  mod_name = mod_name.strip()
  attr_name = attr_name.strip()

  mod_path = module_path(mod_name, root)

  if mod_path is None:
    try:
      mod = importlib.import_module(mod_name)

    except ImportError as e:
      raise EntryPointError(f"failed to load '{entry_point}'") from e

  else:
    try:
      mod = load_module(
        name = mod_name,
        path = mod_path,
        root = root )

    except EntryPointError as e:
      raise

    except Exception as e:
      raise EntryPointError(f"failed to load '{entry_point}'") from e


  if not hasattr( mod, attr_name ):
    raise EntryPointError(
      f"Loaded module '{mod_name}' does not have attribute: '{attr_name}'" )

  func = getattr( mod, attr_name )

  return func

#===============================================================================
class EntryPoint:
  r"""

  Parameters
  ----------
  pyproj : PyProjBase
  root : pathlib.Path
    Root project directory.
  name : str
    A name for this entry point
  logger : logging.Logger
  entry : str
    Entry point spec

    * https://packaging.python.org/en/latest/specifications/entry-points/
  """

  #-----------------------------------------------------------------------------
  def __init__( self,
    pyproj,
    root,
    name,
    logger,
    entry ):

    self.pyproj = pyproj
    self.root = root
    self.name = name
    self.logger = logger
    self.entry = entry

    try:
      self.func = load_entrypoint(
        entry_point = entry,
        root = root )

      logger.info(f"loaded entry-point '{entry}'")

    except Exception as e:
      raise EntryPointError(f"failed to load '{entry}'") from e

  #-----------------------------------------------------------------------------
  def __call__(self, **kwargs):

    cwd = os.getcwd()

    try:

      with validating( file = f"{self.name} -> {self.entry}" ):
        self.func(
          self.pyproj,
          logger = self.logger,
          **kwargs )

    except Exception as e:
      raise EntryPointError(f"failed to run '{self.entry}'") from e

    finally:
      os.chdir(cwd)



from __future__ import annotations
import os
from pathlib import Path
import shutil
import subprocess

from ..pyproj import PyProjBase
from logging import Logger

from ..validate import (
  validating,
  ValidationError,
  ValidPathError,
  FileOutsideRootError )

#===============================================================================
def cargo(
  pyproj: PyProjBase,
  logger: Logger,
  options: dict,
  work_dir,
  src_dir: Path,
  build_dir: Path,
  prefix: Path,
  setup_args: list[str],
  compile_args: list[str],
  install_args: list[str],
  build_clean: bool,
  runner):
  """Run cargo build
  """

  if not shutil.which('cargo'):
    raise ValueError("The 'cargo' program not found.")

  if setup_args or install_args:
    raise ValueError(
      f"cargo builder supports 'compile_args', not 'setup_args' or 'install_args': {setup_args}, {install_args}")

  compile_args = [
    'cargo',
    'build',
    *compile_args,
    '--target-dir',
    str(build_dir) ]

  install_args = [
    'cargo',
    'install',
    *install_args,
    '--target-dir',
    str(build_dir),
    '--root', str(prefix) ]

  runner.run(compile_args)




from __future__ import annotations
import os
import os.path as osp
from pathlib import PurePath
import re
from collections import namedtuple

from .utils import _subdir

#===============================================================================
class PathPatternError(ValueError):
  pass

#===============================================================================
# NOTE: The regular expressions are constructed to match path separators defined
# by these (non-printable) control characters, unlikely to be in any filename,
# making them independent from 'os.path.sep', instead of specializing the patterns
# to each flavor.
# The paths to be matched, though, must be translated to this form.
# File Separator (E.G. '/')
SEP = chr(0x1c)
# Group Separator (E.G. '.')
CURDIR = chr(0x1d)
# Record Separator (E.G. '..')
PARDIR = chr(0x1e)

# NOTE: For debugging, uncomment for printable characters, but may also lead to
# false positive/negative matches depending on the inputs.
# SEP = '/'
# CURDIR = '.'
# PARDIR = '..'

#===============================================================================
def tr_path(path: PurePath) -> str:
  """Translates path to be compatible with the translated regex.match

  Parameters
  ---------
  path:
    Path to be translated

  Returns
  -------
  tr_path:
    Translated path, with each path segment separated by :data:`SEP`.
  """
  parts = path.parts
  anchor = path.anchor

  if not len(parts):
    return ''

  if parts[0] == anchor:
    return SEP.join(parts[1:])

  return SEP.join(parts)

#===============================================================================
def inv_path(path: str, sep: str = osp.sep) -> str:
  """Convert translated path back to os path
  """
  return sep.join(path.split(SEP))

#===============================================================================
# The glob pattern is defined using POSIX paths, even if matching other flavors
# path separator (NOTE: except for a trailing recursive "/**")
re_sep = r'(?P<sep>/(?!\*\*\Z))'
# fixed (no wildcard) segment
re_fixed = r'(?P<fixed>(?:\\[*?[]|[^*?[/])+)'
# single star "*" wildcard (not double star "**") e.g. "*.txt"
re_any = r'(?P<any>(?<![\*\\])\*(?!\*))'
# single character wildcard e.g. "abc_?"
re_chr = r'(?P<chr>(?<!\\)\?)'
# character set e.g. "[a-z]"
re_chrset = r'(?P<chrset>(?<!\\)\[[!^]?\]?[^\]]*\])'
# double star sub-directory e.g. "a/**/b" or "**/b"
# NOTE: the ending '/' is consumed so the replaced pattern also matches zero times
# but leading '/' is not so that zero-length matches leave a '/' for successive
# sub-directory/file patterns.
re_subdir = r'(?P<subdir>(?<=/)\*\*/)'
re_isubdir = r'(?P<isubdir>\A\*\*/)'
# trailing double star e.g. "a/**"
re_alldir = r'(?P<alldir>/\*\*\Z)'

re_glob = '|'.join([
  re_sep,
  re_fixed,
  re_any,
  re_chr,
  re_chrset,
  re_subdir,
  re_isubdir,
  re_alldir ])

rec_glob = re.compile(re_glob)
rec_unescape = re.compile(r'\\([*?[])')

#===============================================================================
def tr_rel_join(start: str, dir: str, names: list[str]) -> list[tuple[str, str]]:
  """Creates paths relative to a 'start' path for a list of names in a 'dir'
  'start' and 'dir' must already be translated by :func:`tr_path`.

  Parameters
  ----------
  start:
    Starting directory, already translated by :func:`tr_path`.
  dir: str
    Directory to compute relative path to, *must* be a sub-directory of `start`,
    already translated by :func:`tr_path`.
  names:
    List of names in `dir`

  Returns
  -------
  List of names joined with path relative to `start`
  """

  rpath = tr_subdir( start, dir )
  #DEBUG print(f"  rpath: {rpath}")

  return [
    (name, tr_join(rpath, name))
    for name in names ]

#===============================================================================
def tr_join(*args: str):
  """Joins paths already translated by :func:`tr_path`.
  """

  args = [x for x in args if x]

  return SEP.join(args)

#===============================================================================
def tr_subdir(start: str, path: str) -> str:
  """Relative path, restricted to sub-directories.

  Parameters
  ----------
  start:
    Starting directory, already translated by :func:`tr_path`.
  path:
    Directory to compute relative path to, *must* be a sub-directory of `start`,
    already translated by :func:`tr_path`.

  Returns
  -------
  rpath:
    Relative path from `start` to `path`.
  """

  #DEBUG print(f"  tr_subdir({start}, {path})")
  if not start:
    return path

  _start = start.split(SEP)
  _path = path.split(SEP)

  _rpath = _subdir(_start, _path)

  if _rpath is None:
    raise PathPatternError(f"Not a subdirectory of {inv_path(start)}: {inv_path(path)}")

  return SEP.join(_rpath)

#===============================================================================
class GRef(namedtuple('GRef', ['ori', 'case', 'start', 'end'])):
  r"""Helps track how a regex was constructed from a glob pattern
  """
  __slots__ = ()


#===============================================================================
class GCase:
  r"""Container for constructing parts of a regex from glob pattern
  """
  #-----------------------------------------------------------------------------
  def __init__(self, ref: GRef|None = None):
    if ref is None:
      ref = GRef(None, 'undefined', None, None)

    self.ref = ref

  #-----------------------------------------------------------------------------
  def regex(self) -> str:
    raise NotImplementedError("")

  #-----------------------------------------------------------------------------
  def __str__(self):
    return self.regex()

#===============================================================================
class GStr(GCase):
  #-----------------------------------------------------------------------------
  def __init__(self, regex: str, ref: GRef|None = None):
    super().__init__(ref = ref)

    self._regex = regex

  #-----------------------------------------------------------------------------
  def regex(self) -> str:
    return self._regex

#===============================================================================
class GList(GCase):
  #-----------------------------------------------------------------------------
  def __init__(self, parts: list[str] = None, ref: GRef|None = None):
    super().__init__(ref = ref)

    if parts is None:
      parts = list()

    self.parts = parts

  #-----------------------------------------------------------------------------
  def regex(self) -> str:
    return ''.join([v.regex() for v in self.parts])

  #-----------------------------------------------------------------------------
  def append(self, val: str):
    self.parts.append(val)

  #-----------------------------------------------------------------------------
  def __str__(self):
    return self._regex

#===============================================================================
class GName(GStr):
  pass

#===============================================================================
class GFixed(GName):
  pass

#===============================================================================
class GChrSet(GName):
  pass

#===============================================================================
GCHR = GName(rf'[^{SEP}]')
GANY = GName(rf'[^{SEP}]*')

#===============================================================================
def reduce_any(pid, sid, n, i, working):
  r"""
  Parameters
  ----------
  pid : int
    Pattern id, may be the same for all segments, but unique among all patterns
    that might be combined.
  sid : int
    Segment id, unique only within an overall pattern.
  n : int
    Number of '*' in the segment `sid`
  i : int
    Current '*' within the segment `sid`
  working : list[GCase]
    Expressions preceeding the '*' at `i`

  Returns
  -------
  regex : str

  Notes
  -----
  Adapted according to:
  bpo-40480: "fnmatch" exponential execution time
  https://github.com/python/cpython/pull/19908
  https://github.com/python/cpython/pull/20049

  Each '*' (any) is grouped with the succeeding non-'*' into a 'lazy'
  non-capturing group.
  Since this match does not depend on the rest of the expressions, it does not
  need to backtrack to account for failures later on.
  A positive lookahead references the match and then asserts it as the
  new pattern.
  Each successive '*' group will absorb anything missed by the previous one
  until the last '*' is reached.
  The last '*' must be allowed to backtrack, since no more '*' exist to make
  up for misses (doesn't really matter if the last is 'greedy' or 'lazy').

  .. code-block:: python

    >>> p = re.compile(r'\A(?=(?P<g1>.*?A))(?P=g1)\Z')
    >>> bool(p.match('xxxA'))
    True
    >>> bool(p.match('xxxAxxxA'))
    False

  This fails because a lazy match to the first 'xxxA' misses the second 'xxxA'.
  For 3 '*', there should be 2 lazy+lookahead, and the last will be unconditional

  .. code-block::

          n: 3
    pattern: ...*...*...*...
          i:    0   1   2   3
    working: 000 111 222 333
     groups:    [  ][  ]

  Groups are named according to ``f'p{pid}s{sid}g{i-1}'``.

  """

  pat = ''.join([v.regex() for v in working])

  if i > 0:
    # This call is actually adding the i-1 group
    # NOTE: i==0 (and n==0) simply returns the preceeding working pattern, since
    # there is not a -1 group

    if i == n:
      # NOTE: i==n means this is the final call, but the n-1 '*' should
      # *not* have a lazy+lookahead group.
      # this also means that for n==1, this will be the only translated '*'.
      pat = rf'{GANY}{pat}'
    else:
      name = f'p{pid}s{sid}g{i-1}'
      pat = rf'(?=(?P<{name}>{GANY}?{pat}))(?P={name})'

  return pat

#===============================================================================
class GSegment(GList):
  #-----------------------------------------------------------------------------
  def __init__(self, pid, sid, parts = None, ref = None):
    super().__init__(parts = parts, ref = ref)
    self.pid = pid
    self.sid = sid

  #-----------------------------------------------------------------------------
  def regex(self):
    if not self.parts:
      return ''

    # collapse repeated '*' and count
    n = int(self.parts[0] is GANY)
    parts = [self.parts[0]]

    for v in self.parts:
      isany = v is GANY

      if not (isany and parts[-1] is GANY):
        parts.append(v)
        n += int(isany)

    # construct groups
    i = 0
    combined = list()
    working = list()

    for v in self.parts:
      if v is GANY:
        combined.append(reduce_any(self.pid, self.sid, n, i, working))
        working = list()
        i += 1
      else:
        working.append(v)

    combined.append(reduce_any(self.pid, self.sid, n, i, working))

    return ''.join(combined)

#===============================================================================
class GSeparator(GStr):
  pass

GSEP = GSeparator(SEP)
GSUBDIR = GSeparator(rf'([^{SEP}]+{SEP})*')
GALLDIR = GSeparator(rf'({SEP}[^{SEP}]+)+')

#===============================================================================
class GPath(list):
  def regex(self):
    return ''.join([v.regex() for v in self])

#===============================================================================
class PatternError(ValueError):
  #-----------------------------------------------------------------------------
  def __init__(self, msg, pat, segs):
    segs = '\n  '.join([ str(seg) for seg in segs])

    msg = f"{msg}: {pat}\n  {segs}"

    super().__init__(msg)
    self.msg = msg
    self.pat = pat
    self.segs = segs

#===============================================================================
def esc_chrset(c):
  if c == '/':
    raise PathPatternError("Path separator '/' in character range is undefined.")

  if c in r'\]-':
    return '\\' + c

  return c

#===============================================================================
def tr_range(pat):

  # range
  a, _, d = pat

  _a = ord(a)
  _d = ord(d)
  _sep = ord('/')

  if _d < _a:
    raise PathPatternError(f"Character range is out of order: {a}-{d} -> {_a}-{_d}")

  if _a <= _sep and _sep <= _d:
    # ranges do not match forward slash '/'
    # E.G. "[--0]" matches the three characters '-', '.', '0'
    b = chr(_sep-1)
    c = chr(_sep+1)

    return ''.join([
      f"{esc_chrset(a)}-{esc_chrset(b)}" if a != b else esc_chrset(a),
      f"{esc_chrset(c)}-{esc_chrset(d)}" if d != c else esc_chrset(d) ])

  return f"{esc_chrset(a)}-{esc_chrset(d)}"

#===============================================================================
def tr_chrset(pat):
  n = len(pat)

  if n <= 2 or pat[0] != '[' or pat[-1] != ']':
    raise PathPatternError(f"Character set must be non-empty: {pat}")

  wild = pat[1:-1]
  parts = ['[']
  add = parts.append

  # NOTE: a lot of this logic rests on the idea that a character class
  # cannot be an empty set. e.g. [] would not match anything, so is not allowed.
  # This means that []] is valid since the the first pair "[]" cannot close
  # a valid set.
  # Likewise, [!] cannot complement an empty set, since this would be equivalent
  # to *, meaning it should instead match "!".
  # [!] -> match "!"
  # [!!] -> match any character that is not "!"
  # []] -> match "]"
  # [!]] -> match any character that is not "]"
  # []!] -> match "]" or "!"

  # NOTE: POSIX has declared the effect of a wildcard pattern "[^...]" to be undefined.
  # Since the standard does not define this behaviour, it seems reasonable to
  # treat "^" the same as "!" due to its common meaning.

  if wild[0] in '!^' and len(wild) > 1:
    # Complement of the set of characters
    # An expression "[!...]" matches a single character, namely any
    # character that is not matched by the expression obtained by
    # removing the first '!' from it.
    add('^')
    wild = wild[1:]

  while wild:
    if len(wild) > 2 and wild[1] == '-':
      # two characters separated by '-' denote a range of characters in the set
      # defined by the ordinal
      add(tr_range(wild[:3]))
      wild = wild[3:]

    else:
      # a single character in the set
      add(esc_chrset(wild[:1]))
      wild = wild[1:]

  parts.append(']')

  return ''.join(parts)

#===============================================================================
def tr_glob(pat, pid = 0):
  """
  Notes
  -----
  * https://man7.org/linux/man-pages/man7/glob.7.html

  """

  # collapse repeated separators '//...' to single '/'
  pat = re.sub(r'/+', '/', pat)

  refs = list()
  segs = GPath()

  def add(case):
    if isinstance(case, GSeparator):
      segs.append(case)
      return

    if not ( len(segs) and isinstance(segs[-1], GSegment) ):
      segs.append(GSegment(pid = pid, sid = len(segs)))

    segs[-1].append(case)


  i = 0

  for m in rec_glob.finditer(pat):

    d = [ k for k,v in m.groupdict().items() if v is not None ]
    assert len(d) == 1

    if i != m.start():
      undefined = pat[i:m.start()]
      refs.append(GRef(undefined, 'undefined', i, m.start()))
      raise PatternError("Invalid pattern", pat, refs)

    refs.append(GRef(m.group(0), d[0], m.start(), m.end()))

    if m['fixed']:
      # NOTE: unescape glob pattern 'escaped' characters before applying re.escape
      # otherwise they become double-escaped
      fixed = rec_unescape.sub(r'\1', m['fixed'])
      add(GFixed(re.escape(fixed)))

    elif m['sep']:
      add(GSEP)

    elif m['subdir'] or m['isubdir']:
      add(GSUBDIR)

    elif m['alldir']:
      add(GALLDIR)

    elif m['any']:
      add(GANY)

    elif m['chr']:
      add(GCHR)

    elif m['chrset']:
      try:
        add(GChrSet(tr_chrset(m['chrset'])))
      except ValueError as e:
        raise PatternError("Invalid pattern", pat, refs) from e

    else:
      assert False, f"Segment case undefined: {m}"

    i = m.end()

  if i != len(pat):
    undefined = pat[i:m.start()]
    refs.append(GRef(undefined, 'undefined', i, len(pat)))
    raise PatternError("Invalid pattern", pat, refs)

  #DEBUG print(segs)
  res = segs.regex()
  return fr'\A{res}\Z', refs
